# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.filelists.filelist import FileList
from gitfourchette.gitdriver import GitDelta
from gitfourchette.globalshortcuts import GlobalShortcuts
from gitfourchette.localization import *
from gitfourchette.nav import NavContext
from gitfourchette.porcelain import *
from gitfourchette.qt import *
from gitfourchette.repomodel import RepoModel
from gitfourchette.tasks import *
from gitfourchette.toolbox import *


class StagedFiles(FileList):
    def __init__(self, repoModel: RepoModel, parent: QWidget):
        super().__init__(repoModel, parent, NavContext.STAGED)

        self.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)

        makeWidgetShortcut(self, self.unstage, *(GlobalShortcuts.discardHotkeys + GlobalShortcuts.stageHotkeys))

    def contextMenuActions(self, deltas: list[GitDelta]) -> list[ActionDef]:
        actions = []

        n = len(deltas)
        modeSet = {delta.new.mode for delta in deltas}  # mode in worktree
        anySubmodules = FileMode.COMMIT in modeSet
        onlySubmodules = anySubmodules and len(modeSet) == 1

        if not anySubmodules:
            contextMenuActionUnstage = ActionDef(
                _n("&Unstage File", "&Unstage {n} Files", n),
                self.unstage,
                icon="git-unstage",
                shortcuts=GlobalShortcuts.discardHotkeys[0])

            actions += [
                contextMenuActionUnstage,
                self.contextMenuActionStash(),
                self.contextMenuActionRevertMode(deltas, self.unstageModeChange, ellipsis=False),
                ActionDef.SEPARATOR,
                self.contextMenuActionBlame(deltas),
                *self.contextMenuActionsDiff(deltas),
                ActionDef.SEPARATOR,
                *self.contextMenuActionsEdit(deltas),
            ]

        elif onlySubmodules:
            actions += [
                ActionDef(
                    _n("Submodule", "{n} Submodules", n),
                    kind=ActionDef.Kind.Section,
                ),

                ActionDef(
                    _n("Unstage Submodule", "Unstage {n} Submodules", n),
                    self.unstage,
                ),

                ActionDef(
                    _n("Open Submodule in New Tab", "Open {n} Submodules in New Tabs", n),
                    self.openSubmoduleTabs,
                ),
            ]

        else:
            sorry = _("Can’t unstage this selection in bulk.") + "\n" + _("Please review the files individually.")
            actions += [
                ActionDef(sorry, enabled=False),
            ]

        actions += super().contextMenuActions(deltas)
        return actions

    def wantStageOrUnstage(self):
        self.unstage()

    def unstage(self):
        deltas = list(self.selectedDeltas())
        UnstageFiles.invoke(self, deltas)

    def unstageModeChange(self):
        deltas = list(self.selectedDeltas())
        UnstageModeChanges.invoke(self, deltas)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

"""
Bridge to gettext translations. Use _, _n, _np, _p to localize English text.
"""

from gettext import GNUTranslations
from gettext import NullTranslations


_translator = NullTranslations()


def installGettextTranslator(path: str = "") -> bool:
    """
    Load translations from a gettext '.mo' file.

    Return True if the translations were successfully loaded.

    If the given path is empty or doesn't exist, fall back to
    American English and return False.
    """

    global _translator

    if path:
        try:
            with open(path, 'rb') as fp:
                _translator = GNUTranslations(fp)
                return True
        except OSError:
            pass

    _translator = NullTranslations()
    return False


def _(message: str, *args, **kwargs) -> str:
    message = _translator.gettext(message)
    if args or kwargs:
        message = message.format(*args, **kwargs)
    return message


def _n(singular: str, plural: str, n: int, *args, **kwargs) -> str:
    return _translator.ngettext(singular, plural, n).format(*args, **kwargs, n=n)


def _np(context: str, singular: str, plural: str, n: int) -> str:
    return _translator.npgettext(context, singular, plural, n).format(n=n)


def _p(context: str, message: str, *args, **kwargs) -> str:
    message = _translator.pgettext(context, message)
    if args or kwargs:
        message = message.format(*args, **kwargs)
    return message


__all__ = [
    "_",
    "_n",
    "_np",
    "_p",
    "installGettextTranslator",
]



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

"""
Remote management tasks.
"""

from gitfourchette.forms.remotedialog import RemoteDialog
from gitfourchette.gitdriver import GitDriver, argsIf
from gitfourchette.localization import *
from gitfourchette.porcelain import GitConfigHelper
from gitfourchette.qt import *
from gitfourchette.tasks.repotask import RepoTask, TaskEffects
from gitfourchette.toolbox import *


class NewRemote(RepoTask):
    def flow(self):
        existingRemotes = [r.name for r in self.repo.remotes]

        dlg = RemoteDialog(
            editExistingRemote=False,
            name="",
            url="",
            existingRemotes=existingRemotes,
            skipFetchAll=False,
            parent=self.parentWidget())

        dlg.setWindowModality(Qt.WindowModality.WindowModal)
        dlg.show()
        dlg.setMaximumHeight(dlg.height())
        yield from self.flowDialog(dlg)

        newRemoteName = dlg.ui.nameEdit.text()
        newRemoteUrl = dlg.ui.urlEdit.text()
        fetchAfterAdd = dlg.ui.fetchAfterAddCheckBox.isChecked()
        newSkipFetchAll = dlg.ui.skipFetchAllCheckBox.isChecked()
        dlg.deleteLater()

        self.epilog.effects |= TaskEffects.Refs | TaskEffects.Remotes
        yield from self.flowCallGit("remote", "add", "--", newRemoteName, newRemoteUrl)
        self.repo.set_remote_skipfetchall(newRemoteName, newSkipFetchAll)

        self.epilog.status = _("Remote {0} added.", tquo(newRemoteName))

        if fetchAfterAdd:
            from gitfourchette.tasks import FetchRemotes
            yield from self.flowSubtask(FetchRemotes, newRemoteName)


class EditRemote(RepoTask):
    def flow(self, oldRemoteName: str):
        oldRemoteUrl = self.repo.remotes[oldRemoteName].url

        existingRemotes = [r.name for r in self.repo.remotes]
        existingRemotes.remove(oldRemoteName)

        skipFetchAllConfigKey = GitConfigHelper.sanitize_key(("remote", oldRemoteName, "skipFetchAll"))
        try:
            skipFetchAll = self.repo.config.get_bool(skipFetchAllConfigKey)
        except KeyError:
            skipFetchAll = False

        dlg = RemoteDialog(
            editExistingRemote=True,
            name=oldRemoteName,
            url=oldRemoteUrl,
            existingRemotes=existingRemotes,
            skipFetchAll=skipFetchAll,
            parent=self.parentWidget())

        dlg.setWindowModality(Qt.WindowModality.WindowModal)
        dlg.show()
        yield from self.flowDialog(dlg)

        newRemoteName = dlg.ui.nameEdit.text()
        newRemoteUrl = dlg.ui.urlEdit.text().strip()
        newSkipFetchAll = dlg.ui.skipFetchAllCheckBox.isChecked()
        dlg.deleteLater()

        self.epilog.effects |= TaskEffects.Refs | TaskEffects.Remotes

        if oldRemoteName != newRemoteName:
            yield from self.flowCallGit(
                "remote",
                "rename",
                *argsIf(GitDriver.supportsDashDashBeforePositionalArgs(), "--progress", "--"),
                oldRemoteName,
                newRemoteName)

        if oldRemoteUrl != newRemoteUrl:
            yield from self.flowCallGit(
                "remote",
                "set-url",
                "--",
                newRemoteName,
                newRemoteUrl)

        self.repo.set_remote_skipfetchall(newRemoteName, newSkipFetchAll)

        self.epilog.status = _("Remote {0} modified.", tquo(newRemoteName))


class DeleteRemote(RepoTask):
    def flow(self, remoteName: str):
        yield from self.flowConfirm(
            text=paragraphs(
                _("Really remove remote {0}?", bquo(remoteName)),
                _("This will merely detach the remote from your local repository. "
                  "The remote server itself will not be affected.")),
            verb=_("Remove remote"),
            buttonIcon="SP_DialogDiscardButton")

        self.epilog.effects |= TaskEffects.Refs | TaskEffects.Remotes

        yield from self.flowCallGit(
            "remote",
            "remove",
            *argsIf(GitDriver.supportsDashDashBeforePositionalArgs(), "--"),
            remoteName)

        self.epilog.status = _("Remote {0} removed.", tquo(remoteName))



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.blameview.blamegutter import BlameGutter
from gitfourchette.blameview.blamemodel import BlameModel, Revision
from gitfourchette.codeview.codeview import CodeView
from gitfourchette.localization import *
from gitfourchette.nav import NavLocator
from gitfourchette.qt import *
from gitfourchette.tasks import TaskBook, GetCommitInfo
from gitfourchette.toolbox import *


class BlameTextEdit(CodeView):
    showRevision = Signal(Revision)
    jumpToCommit = Signal(NavLocator)

    model: BlameModel
    gutter: BlameGutter

    def __init__(self, model, parent=None):
        super().__init__(gutterClass=BlameGutter, parent=parent)
        self.model = model
        self.gutter.model = model

    def contextMenuActions(self, clickedCursor: QTextCursor):
        lineNumber = clickedCursor.blockNumber() + 1
        lineNumber = min(lineNumber, len(self.model.currentRevision.blameLines) - 1)

        commitId = self.model.currentRevision.blameLines[lineNumber].commitId
        try:
            revision = self.model.revList.revisionForCommit(commitId)
            locator = revision.toLocator()
        except LookupError:
            revision = None
            locator = NavLocator.inCommit(commitId)
        isWorkdir = locator.context.isWorkdir()

        if isWorkdir:
            blameLabel = _("Blame File at Uncommitted Revision")
            gotoLabel = _("Show Diff in Working Directory")
        else:
            blameLabel = _("Blame File at {0}", tquo(shortHash(commitId)))
            gotoLabel = _("Show {0} in Repo", tquo(shortHash(commitId)))

        return [
            ActionDef(
                blameLabel,
                icon="git-blame",
                callback=lambda: self.showRevision.emit(revision),
                enabled=revision is not None,
            ),

            ActionDef(
                gotoLabel,
                icon="go-window",
                callback=lambda: self.jumpToCommit.emit(locator)
            ),

            TaskBook.action(
                self,
                GetCommitInfo,
                taskArgs=[commitId, False, self.onGetInfoLocator],
                icon="SP_MessageBoxInformation",
                enabled=not isWorkdir,
            ),
        ]

    def onGetInfoLocator(self, locator: NavLocator):
        self.jumpToCommit.emit(locator)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import logging

from gitfourchette.appconsts import *
from gitfourchette.codeview.codehighlighter import CodeHighlighter
from gitfourchette.diffview.diffdocument import DiffDocument, LineData
from gitfourchette.syntax import LexJob, ColorScheme

logger = logging.getLogger(__name__)


class DiffHighlighter(CodeHighlighter):
    oldLexJob: LexJob | None
    newLexJob: LexJob | None

    def __init__(self, parent):
        super().__init__(parent)
        self.diffDocument = None
        self.oldLexJob = None
        self.newLexJob = None

    def setDiffDocument(self, diffDocument: DiffDocument):
        self.diffDocument = diffDocument
        self.setDocument(diffDocument.document)

        # Prime lex jobs
        self.stopLexJobs()
        self.oldLexJob = diffDocument.oldLexJob
        self.newLexJob = diffDocument.newLexJob
        for job in self.oldLexJob, self.newLexJob:
            if job is not None:
                self.installLexJob(job)

    def highlightSyntax(self, text: str):
        # Pygments syntax highlighting
        blockNumber = self.currentBlock().blockNumber()

        lineData: LineData = self.diffDocument.lineData[blockNumber]
        if not lineData.origin:  # Hunk header, etc.
            return

        if lineData.origin != '-':
            lexJob = self.newLexJob
            lineNumber = lineData.newLineNo
        else:
            lexJob = self.oldLexJob
            lineNumber = lineData.oldLineNo

        if lexJob is None:
            return

        column = 0
        scheme = self.scheme.highContrastScheme if lineData.origin in "+-" else self.scheme.scheme
        boundary = len(text) - lineData.trailerLength

        for tokenType, tokenLength in lexJob.tokens(lineNumber, text):
            try:
                charFormat = scheme[tokenType]
            except KeyError:
                charFormat = ColorScheme.fillInFallback(scheme, tokenType)
            self.setFormat(column, tokenLength, charFormat)
            column += tokenLength
            if column >= boundary:
                break

        if APP_DEBUG and lexJob.lexingComplete and column != boundary:  # pragma: no cover
            # Overstep may occur in low-quality lexing (not a big deal, so
            # we ignore that case) or when the file isn't decoded properly.
            logger.warning(f"Syntax highlighting overstep on line -{lineData.oldLineNo}+{lineData.newLineNo} {column} != {boundary}")



# -----------------------------------------------------------------------------
# Copyright (C) 2024 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from collections.abc import Iterable

from gitfourchette.graph.graph import Oid

STOP = 0
PIPE = 1
SOURCE = 2


class GraphTrickle:
    def __init__(self):
        self.frontier = {}
        self.flaggedSet = set()

    @property
    def done(self) -> bool:
        return all(v == STOP for v in self.frontier.values())

    def newCommit(self, commit: Oid, parents: Iterable[Oid]):
        frontier = self.frontier
        flagged = frontier.pop(commit, STOP)

        if flagged != STOP:
            # Trickle through parents that are not explicitly flagged
            for p in parents:
                frontier.setdefault(p, PIPE)
            self.flaggedSet.add(commit)
        else:
            # Block trickling to parents (that aren't sources themselves)
            for p in parents:
                if frontier.get(p, STOP) != SOURCE:
                    frontier[p] = STOP

    @staticmethod
    def newHiddenTrickle(
            allHeads: set[Oid],
            hideSeeds: set[Oid],
            forceHide: set[Oid] | None = None
    ):
        trickle = GraphTrickle()

        # Explicitly show all refs by default (block foreign trickle)
        for head in allHeads:
            trickle.frontier[head] = STOP

        # Explicitly hide tips (allow foreign trickle)
        for head in hideSeeds:
            trickle.frontier[head] = PIPE

        # Explicitly hide stash junk parents (beyond parent #0)
        # NOTE: Dropped this from the actual app but kept around in unit tests.
        if forceHide:
            for head in forceHide:
                trickle.frontier[head] = SOURCE

        assert trickle.testFrontierInputs()
        return trickle

    @staticmethod
    def newForeignTrickle(
            allHeads: set[Oid],
            localSeeds: set[Oid]
    ):
        trickle = GraphTrickle()

        # Start with all foreign heads
        for head in allHeads:
            trickle.frontier[head] = PIPE

        # Local heads block propagation of foreign trickle
        for head in localSeeds:
            trickle.frontier[head] = STOP

        assert trickle.testFrontierInputs()
        return trickle

    def testFrontierInputs(self):
        return all(isinstance(head, Oid) for head in self.frontier)



# -----------------------------------------------------------------------------
# Copyright (C) 2025 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.forms.ui_registersubmoduledialog import Ui_RegisterSubmoduleDialog
from gitfourchette.localization import *
from gitfourchette.qt import *
from gitfourchette.toolbox import *


class RegisterSubmoduleDialog(QDialog):
    def __init__(
            self,
            workdirPath: str,
            superprojectName: str,
            remotes: dict[str, str],
            absorb: bool,
            reservedNames: set[str],
            parent):
        super().__init__(parent)
        ui = Ui_RegisterSubmoduleDialog()
        ui.setupUi(self)
        self.ui = ui
        self.reservedNames = reservedNames

        for k, v in remotes.items():
            ui.remoteComboBox.addItemWithPreview(k, v, v)

        ui.pathValue.setText(workdirPath)
        ui.nameEdit.setText(workdirPath)
        ui.nameEdit.addAction(stockIcon("git-submodule"), QLineEdit.ActionPosition.LeadingPosition)
        self.resetNameAction = ui.nameEdit.addAction(stockIcon("SP_LineEditClearButton"), QLineEdit.ActionPosition.TrailingPosition)
        self.resetNameAction.setVisible(False)
        self.resetNameAction.setToolTip(_("Reset to default name"))
        self.resetNameAction.triggered.connect(lambda: ui.nameEdit.setText(workdirPath))
        ui.nameEdit.textChanged.connect(lambda name: self.resetNameAction.setVisible(name != workdirPath))

        if absorb:
            formatWidgetText(ui.absorbExplainer, sub=lquoe(workdirPath), super=lquoe(superprojectName))
            self.okButton.setText(_("Absorb submodule"))
            self.setWindowTitle(_("Absorb submodule"))
        else:
            ui.absorbExplainer.hide()
            self.okButton.setText(_("Register submodule"))
            self.setWindowTitle(_("Register submodule"))

        validator = ValidatorMultiplexer(self)
        validator.setGatedWidgets(self.okButton)
        validator.connectInput(self.ui.nameEdit, self.validateSubmoduleName)
        validator.run()

    @property
    def remoteUrl(self) -> str:
        return self.ui.remoteComboBox.currentData(Qt.ItemDataRole.UserRole)

    @property
    def customName(self):
        return self.ui.nameEdit.text()

    @property
    def okButton(self):
        return self.ui.buttonBox.button(QDialogButtonBox.StandardButton.Ok)

    def validateSubmoduleName(self, name: str):
        if not name.strip():
            return _("Cannot be empty.")
        elif name in self.reservedNames:
            return _("This name is already taken by another submodule.")
        else:
            return ""



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import logging
from contextlib import suppress
from pathlib import Path

import pygit2

from gitfourchette.forms.brandeddialog import convertToBrandedDialog
from gitfourchette.forms.checkoutcommitdialog import CheckoutCommitDialog
from gitfourchette.forms.commitdialog import CommitDialog
from gitfourchette.forms.deletetagdialog import DeleteTagDialog
from gitfourchette.forms.identitydialog import IdentityDialog
from gitfourchette.forms.newtagdialog import NewTagDialog
from gitfourchette.forms.signatureform import SignatureOverride
from gitfourchette.gitdriver import argsIf
from gitfourchette.localization import *
from gitfourchette.nav import NavLocator
from gitfourchette.porcelain import *
from gitfourchette.qt import *
from gitfourchette.repomodel import GpgStatus
from gitfourchette.tasks.jumptasks import RefreshRepo
from gitfourchette.tasks.repotask import AbortTask, RepoTask, TaskPrereqs, TaskEffects
from gitfourchette.toolbox import *

logger = logging.getLogger(__name__)


class NewCommit(RepoTask):
    def prereqs(self):
        return TaskPrereqs.NoConflicts

    def flow(self):
        from gitfourchette.tasks import Jump

        uiPrefs = self.repoModel.prefs

        # Jump to workdir
        yield from self.flowSubtask(Jump, NavLocator.inWorkdir())

        emptyCommit = not self.repo.any_staged_changes
        if emptyCommit:
            text = [_("No files are staged for commit."), _("Do you want to create an empty commit anyway?")]

            if self.repoModel.numUncommittedChanges != 0:
                text.append("<small>" + _n(
                    "Note: Your working directory contains {n} unstaged file. "
                    "If you want to commit it, you should <b>stage</b> it first.",
                    "Note: Your working directory contains {n} unstaged files. "
                    "If you want to commit them, you should <b>stage</b> them first.",
                    self.repoModel.numUncommittedChanges) + "</small>")

            yield from self.flowConfirm(
                title=_("Create empty commit"),
                verb=_("Empty commit"),
                text=paragraphs(*text))

        yield from self.flowSubtask(SetUpGitIdentity, self.name())

        repositoryState = self.repo.state()
        fallbackSignature = self.repo.default_signature
        initialMessage = uiPrefs.draftCommitMessage
        gpgFlag, gpgKey = NewCommit.getGpgConfig(self.repo)

        cd = CommitDialog(
            initialText=initialMessage,
            authorSignature=fallbackSignature,
            committerSignature=fallbackSignature,
            amendingCommitHash="",
            detachedHead=self.repo.head_is_detached,
            repositoryState=repositoryState,
            emptyCommit=emptyCommit,
            gpgFlag=gpgFlag,
            gpgKey=gpgKey,
            hooks=self.preCommitHookNames(self.repo),
            parent=self.parentWidget())

        if uiPrefs.draftCommitSignatureOverride == SignatureOverride.Nothing:
            cd.ui.revealSignature.setChecked(False)
        else:
            assert uiPrefs.draftCommitSignature is not None, "overridden Signature can't be None"
            cd.ui.revealSignature.setChecked(True)
            cd.ui.signature.setSignature(uiPrefs.draftCommitSignature)
            cd.ui.signature.ui.replaceComboBox.setCurrentIndex(int(uiPrefs.draftCommitSignatureOverride) - 1)

        cd.setWindowModality(Qt.WindowModality.WindowModal)

        # Reenter task even if dialog rejected, because we want to save the commit message as a draft
        yield from self.flowDialog(cd, abortTaskIfRejected=False)

        message = cd.getFullMessage()
        author = cd.getOverriddenAuthorSignature() or fallbackSignature
        committer = cd.getOverriddenCommitterSignature() or fallbackSignature
        overriddenSignatureKind = cd.getOverriddenSignatureKind()
        signatureIsOverridden = overriddenSignatureKind != SignatureOverride.Nothing
        explicitGpgSign = cd.ui.gpg.explicitSign()
        explicitNoGpgSign = cd.ui.gpg.explicitNoSign()
        explicitNoVerify = cd.ui.hookButton.explicitNoVerify()
        signoff = cd.ui.signoffButton.explicitSign()

        # Save commit message/signature as draft now,
        # so we don't lose it if the commit operation fails or is rejected.
        if message != initialMessage or signatureIsOverridden:
            uiPrefs.draftCommitMessage = message
            uiPrefs.draftCommitSignature = cd.ui.signature.getSignature() if signatureIsOverridden else None
            uiPrefs.draftCommitSignatureOverride = overriddenSignatureKind
            uiPrefs.setDirty()

        cd.deleteLater()

        if cd.result() == QDialog.DialogCode.Rejected:
            raise AbortTask()

        self.epilog.effects |= TaskEffects.Workdir | TaskEffects.Refs | TaskEffects.Head
        args, env = NewCommit.prepareGitCommand(
            message, author, committer,
            repositoryState=repositoryState,
            explicitGpgSign=explicitGpgSign,
            explicitNoGpgSign=explicitNoGpgSign,
            explicitNoVerify=explicitNoVerify,
            signoff=signoff)
        driver = yield from self.flowCallGit(*args, env=env)

        branchName, newHash = driver.readPostCommitInfo()
        newOid = Oid(hex=newHash)

        # Trust this commit if we've just signed it
        if explicitGpgSign:
            self.repoModel.cacheGpgStatus(newOid, GpgStatus.GoodTrusted, gpgKey)

        uiPrefs.clearDraftCommit()

        self.epilog.status = _("Commit {0} created on {1}.", tquo(shortHash(newHash)), branchName)

    @staticmethod
    def getGpgConfig(repo: Repo) -> tuple[bool, str]:
        gpgFlag, gpgKey = False, ""
        with suppress(KeyError):
            gpgFlag = repo.config.get_bool("commit.gpgSign")
        with suppress(KeyError):
            gpgKey = repo.config["user.signingKey"]
        return gpgFlag, gpgKey

    @staticmethod
    def preCommitHookNames(repo: Repo) -> list[str]:
        return [hook for hook in ("pre-commit", "commit-msg")
                if Path(repo.in_gitdir("hooks/" + hook)).exists()]

    @staticmethod
    def prepareGitCommand(
            message: str,
            author: Signature | None,
            committer: Signature | None,
            repositoryState: RepositoryState,
            amend=False,
            explicitGpgSign=False,
            explicitNoGpgSign=False,
            explicitNoVerify=False,
            signoff=False,
    ):
        def signatureEnvironmentVariables(sig: Signature, infix: str) -> dict[str, str]:
            return {
                f"GIT_{infix}_NAME": sig.name,
                f"GIT_{infix}_EMAIL": sig.email,
                f"GIT_{infix}_DATE": f"{sig.time}{formatTimeOffset(sig.offset)}",
            }

        # Git ignores GIT_AUTHOR_* when amending or concluding a cherrypick
        # unless we pass --reset-author.
        resetAuthor = bool(author and (amend or repositoryState == RepositoryState.CHERRYPICK))

        args = [
            "-c", "core.abbrev=no",
            "commit",
            *argsIf(explicitGpgSign, "--gpg-sign"),
            *argsIf(explicitNoGpgSign, "--no-gpg-sign"),
            *argsIf(explicitNoVerify, "--no-verify"),
            *argsIf(signoff, "--signoff"),
            *argsIf(amend, "--amend"),
            *argsIf(resetAuthor, "--reset-author"),
            "--allow-empty",
            "--no-edit",
            f"--message={message}"
        ]

        env: dict[str, str] = {}

        if author is not None:
            env |= signatureEnvironmentVariables(author, "AUTHOR")

        if committer is not None:
            env |= signatureEnvironmentVariables(committer, "COMMITTER")

        return args, env


class AmendCommit(RepoTask):
    def prereqs(self):
        return TaskPrereqs.NoUnborn | TaskPrereqs.NoConflicts | TaskPrereqs.NoCherrypick

    def getDraftMessage(self):
        return self.repoModel.prefs.draftAmendMessage

    def setDraftMessage(self, newMessage):
        self.repoModel.prefs.draftAmendMessage = newMessage
        self.repoModel.prefs.setDirty()

    def flow(self):
        from gitfourchette.tasks import Jump

        # Jump to workdir
        yield from self.flowSubtask(Jump, NavLocator.inWorkdir())

        yield from self.flowSubtask(SetUpGitIdentity, self.name())

        repositoryState = self.repo.state()
        headCommit = self.repo.head_commit
        fallbackSignature = self.repo.default_signature
        gpgFlag, gpgKey = NewCommit.getGpgConfig(self.repo)
        emptyCommit = not self.repo.any_staged_changes

        # TODO: Retrieve draft message
        cd = CommitDialog(
            initialText=headCommit.message,
            authorSignature=headCommit.author,
            committerSignature=fallbackSignature,
            amendingCommitHash=shortHash(headCommit.id),
            detachedHead=self.repo.head_is_detached,
            repositoryState=self.repo.state(),
            emptyCommit=emptyCommit,
            gpgFlag=gpgFlag,
            gpgKey=gpgKey,
            hooks=NewCommit.preCommitHookNames(self.repo),
            parent=self.parentWidget())

        cd.setWindowModality(Qt.WindowModality.WindowModal)

        # Reenter task even if dialog rejected, because we want to save the commit message as a draft
        yield from self.flowDialog(cd, abortTaskIfRejected=False)
        cd.deleteLater()

        message = cd.getFullMessage()

        # Save amend message as draft now, so we don't lose it if the commit operation fails or is rejected.
        self.setDraftMessage(message)

        if cd.result() == QDialog.DialogCode.Rejected:
            raise AbortTask()

        author = cd.getOverriddenAuthorSignature()  # no "or fallback" here - leave author intact for amending
        committer = cd.getOverriddenCommitterSignature() or fallbackSignature
        explicitGpgSign = cd.ui.gpg.explicitSign()
        explicitNoGpgSign = cd.ui.gpg.explicitNoSign()
        explicitNoVerify = cd.ui.hookButton.explicitNoVerify()
        signoff = cd.ui.signoffButton.explicitSign()

        self.epilog.effects |= TaskEffects.Workdir | TaskEffects.Refs | TaskEffects.Head
        args, env = NewCommit.prepareGitCommand(
            message, author, committer,
            repositoryState=repositoryState,
            amend=True,
            explicitGpgSign=explicitGpgSign,
            explicitNoGpgSign=explicitNoGpgSign,
            explicitNoVerify=explicitNoVerify,
            signoff=signoff)
        driver = yield from self.flowCallGit(*args, env=env)

        _branchName, newHash = driver.readPostCommitInfo()
        newOid = Oid(hex=newHash)

        # Trust this commit if we've just signed it
        if explicitGpgSign:
            self.repoModel.cacheGpgStatus(newOid, GpgStatus.GoodTrusted, gpgKey)

        self.repoModel.prefs.clearDraftAmend()

        self.epilog.status = _("Commit {0} amended. New hash: {1}.",
                               tquo(shortHash(headCommit.id)),
                               tquo(shortHash(newHash)))


class SetUpGitIdentity(RepoTask):
    def flow(self, okButtonText="", firstRun=True):
        if firstRun:
            # Getting the default signature will fail if the user's identity is missing or incorrectly set
            try:
                _dummy = self.repo.default_signature
                return
            except (KeyError, ValueError):
                pass

        initialName, initialEmail, editLevel = GitConfigHelper.global_identity()

        # Fall back to a sensible path if the identity comes from /etc/gitconfig or some other systemwide file
        if editLevel not in [GitConfigLevel.XDG, GitConfigLevel.GLOBAL]:
            # Favor XDG path if we can, otherwise use ~/.gitconfig
            if FREEDESKTOP and pygit2.settings.search_path[GitConfigLevel.XDG]:
                editLevel = GitConfigLevel.XDG
            else:
                editLevel = GitConfigLevel.GLOBAL

        editPath = GitConfigHelper.path_for_level(editLevel, missing_dir_ok=True)

        dlg = IdentityDialog(firstRun, initialName, initialEmail, editPath,
                             self.repo.has_local_identity(), self.parentWidget())

        if okButtonText:
            proceed = _("Proceed to {0}", englishTitleCase(okButtonText))
            dlg.ui.buttonBox.button(QDialogButtonBox.StandardButton.Ok).setText(proceed)

        dlg.setWindowModality(Qt.WindowModality.WindowModal)
        yield from self.flowDialog(dlg)

        name, email = dlg.identity()
        dlg.deleteLater()

        configObject = GitConfigHelper.ensure_file(editLevel)
        configObject['user.name'] = name
        configObject['user.email'] = email

        # An existing repo will automatically pick up the new GLOBAL config file,
        # but apparently not the XDG config file... So add it to be sure.
        with suppress(ValueError):
            self.repo.config.add_file(editPath, editLevel, force=False)


class CheckoutCommit(RepoTask):
    def prereqs(self) -> TaskPrereqs:
        return TaskPrereqs.NoConflicts

    def flow(self, oid: Oid):
        from gitfourchette.tasks.branchtasks import SwitchBranch, NewBranchFromCommit, ResetHead, MergeBranch

        refs = self.repo.listall_refs_pointing_at(oid)
        refs = [r for r in refs if r.startswith((RefPrefix.HEADS, RefPrefix.REMOTES))]

        commitMessage = self.repo.get_commit_message(oid)
        commitMessage, _junk = messageSummary(commitMessage)
        anySubmodules = bool(self.repo.listall_submodules_fast())

        dlg = CheckoutCommitDialog(
            oid=oid,
            refs=refs,
            currentBranch=self.repo.head_branch_shorthand,
            anySubmodules=anySubmodules,
            parent=self.parentWidget())

        convertToBrandedDialog(dlg, subtitleText=tquo(commitMessage))
        dlg.setWindowModality(Qt.WindowModality.WindowModal)
        yield from self.flowDialog(dlg)

        # Make sure to copy user input from dialog UI *before* starting worker thread
        dlg.deleteLater()

        wantSubmodules = anySubmodules and dlg.ui.recurseSubmodulesCheckBox.isChecked()

        self.epilog.effects |= TaskEffects.Refs | TaskEffects.Head

        if dlg.ui.detachHeadRadioButton.isChecked():
            headId = self.repoModel.headCommitId
            if self.repoModel.dangerouslyDetachedHead() and oid != headId:
                text = paragraphs(
                    _("You are in <b>Detached HEAD</b> mode at commit {0}.", btag(shortHash(headId))),
                    _("You might lose track of this commit "
                      "if you carry on checking out another commit ({0}).", shortHash(oid)))
                yield from self.flowConfirm(text=text, icon='warning')

            yield from self.flowCallGit(
                "checkout",
                "--progress",
                "--detach",
                *argsIf(wantSubmodules, "--recurse-submodules"),
                str(oid))

            self.epilog.status = _("Entered detached HEAD on {0}.", lquo(shortHash(oid)))

            # Force sidebar to select detached HEAD
            self.epilog.jumpTo = NavLocator.inRef("HEAD")

        elif dlg.ui.switchRadioButton.isChecked():
            branchName = dlg.ui.switchComboBox.currentText()
            yield from self.flowSubtask(SwitchBranch, branchName, askForConfirmation=False, recurseSubmodules=wantSubmodules)

        elif dlg.ui.createBranchRadioButton.isChecked():
            yield from self.flowSubtask(NewBranchFromCommit, oid)

        elif dlg.ui.resetHeadRadioButton.isChecked():
            yield from self.flowSubtask(ResetHead, oid)

        elif dlg.ui.mergeRadioButton.isChecked():
            yield from self.flowSubtask(MergeBranch, refs[0])

        else:
            raise NotImplementedError("Unsupported CheckoutCommitDialog outcome")


class NewTag(RepoTask):
    def prereqs(self):
        return TaskPrereqs.NoUnborn

    def flow(self, oid: Oid = NULL_OID, annotation: str | None = None):
        if annotation is not None:
            yield from self.flowSubtask(SetUpGitIdentity, self.name())

        repo = self.repo
        if oid is None or oid == NULL_OID:
            oid = repo.head_commit_id

        reservedNames = repo.listall_tags()
        commitMessage = repo.get_commit_message(oid)
        commitMessage, _dummy = messageSummary(commitMessage)

        dlg = NewTagDialog(shortHash(oid), commitMessage, reservedNames,
                           remotes=self.repoModel.remotes,
                           parent=self.parentWidget())

        yield from self.flowDialog(dlg)

        tagName = dlg.ui.nameEdit.text()
        pushIt = dlg.ui.pushCheckBox.isChecked()
        pushTo = dlg.ui.remoteComboBox.currentData()
        forceCreate = dlg.ui.forceCheckBox.isEnabled() and dlg.ui.forceCheckBox.isChecked()
        dlg.deleteLater()

        self.epilog.effects |= TaskEffects.Refs

        refName = RefPrefix.TAGS + tagName

        if annotation is not None:
            # Annotated tag
            raise NotImplementedError("annotated tags not supported yet")  # git tag -mMSG NAME OID
        else:
            # Lightweight tag
            yield from self.flowCallGit(
                "tag",
                *argsIf(forceCreate, "--force"),
                "--",
                tagName,
                str(oid))

        self.epilog.status = _("Tag {0} created on commit {1}.", tquo(tagName), tquo(shortHash(oid)))

        if pushIt:
            from gitfourchette.tasks import PushRefspecs
            yield from self.flowEnterUiThread()
            yield from self.flowSubtask(PushRefspecs, pushTo, [refName])


class DeleteTag(RepoTask):
    def flow(self, tagName: str):
        assert not tagName.startswith("refs/")

        tagTarget = self.repo.commit_id_from_tag_name(tagName)
        commitMessage = self.repo.get_commit_message(tagTarget)
        commitMessage, _dummy = messageSummary(commitMessage)

        dlg = DeleteTagDialog(
            tagName,
            shortHash(tagTarget),
            commitMessage,
            self.repoModel.remotes,
            parent=self.parentWidget())

        yield from self.flowDialog(dlg)

        pushIt = dlg.ui.pushCheckBox.isChecked()
        pushTo = dlg.ui.remoteComboBox.currentData()
        dlg.deleteLater()

        self.epilog.effects |= TaskEffects.Refs

        # Stay on this commit after the operation
        if tagTarget:
            self.epilog.jumpTo = NavLocator.inCommit(tagTarget)

        yield from self.flowCallGit("tag", "--delete", "--", tagName)

        if pushIt:
            refspec = f":{RefPrefix.TAGS}{tagName}"
            from gitfourchette.tasks import PushRefspecs
            yield from self.flowSubtask(PushRefspecs, pushTo, [refspec])


class RevertCommit(RepoTask):
    def prereqs(self) -> TaskPrereqs:
        return TaskPrereqs.NoConflicts | TaskPrereqs.NoStagedChanges

    def flow(self, oid: Oid):
        text = paragraphs(
            _("Do you want to revert commit {0}?", btag(shortHash(oid))),
            _("You will have an opportunity to review the affected files in your working directory."))
        yield from self.flowConfirm(text=text)

        repoModel = self.repoModel
        repo = self.repo

        self.epilog.effects |= TaskEffects.Workdir

        driver = yield from self.flowCallGit("revert", "--no-commit", "--no-edit", str(oid), autoFail=False)

        # 0: Successful (or dud)
        # 1: Conflicts
        # 128: Local changes would be overwritten, etc.
        exitCode = driver.exitCode()
        if exitCode not in [0, 1]:
            raise AbortTask(driver.htmlErrorText(), details=driver.formatCommandLine())

        # Refresh libgit2 index for conflict analysis
        yield from self.flowEnterWorkerThread()
        self.repo.refresh_index()

        anyConflicts = exitCode != 0
        dud = not anyConflicts and not repo.any_staged_changes

        # If reverting didn't do anything, don't let the REVERT state linger.
        # (Otherwise, the state will be cleared when we commit)
        if dud:
            repo.state_cleanup()

        # Back to UI thread
        yield from self.flowEnterUiThread()

        if dud:
            info = _("There’s nothing to revert from {0} "
                     "that the current branch hasn’t already undone.", bquo(shortHash(oid)))
            raise AbortTask(info, "information")

        repoModel.prefs.draftCommitMessage = self.repo.message_without_conflict_comments
        repoModel.prefs.setDirty()

        self.epilog.jumpTo = NavLocator.inWorkdir()

        if not anyConflicts:
            yield from self.flowSubtask(RefreshRepo, TaskEffects.Workdir, NavLocator.inStaged(""))
            text = _("Reverting {0} was successful. Do you want to commit the result now?", bquo(shortHash(oid)))
            yield from self.flowConfirm(text=text, verb=_p("verb", "Commit"), cancelText=_("Review changes"))
            yield from self.flowSubtask(NewCommit)


class CherrypickCommit(RepoTask):
    def prereqs(self):
        # Prevent cherry-picking with staged changes, like vanilla git (despite libgit2 allowing it)
        return TaskPrereqs.NoConflicts | TaskPrereqs.NoStagedChanges

    def flow(self, oid: Oid):
        question = _("Do you want to apply the changes from commit {0} "
                     "to your working directory?", bquo(shortHash(oid)))
        yield from self.flowConfirm(_("Cherry-pick"), question)

        commit = self.repo.peel_commit(oid)

        self.epilog.effects |= TaskEffects.Workdir
        driver = yield from self.flowCallGit("cherry-pick", "--no-commit", str(oid), autoFail=False)

        exitCode = driver.exitCode()
        if exitCode not in [0, 1]:
            # Surface Git's own error message instead of an opaque exception
            # (e.g. exit code 128 when local changes would be overwritten).
            raise AbortTask(driver.htmlErrorText(), details=driver.formatCommandLine())

        yield from self.flowEnterWorkerThread()

        # We're not going to create a commit right away so the user has a chance
        # to review the changes in the UI. So, force the cherry-picking state.
        # (Note: CHERRY_PICK_HEAD is private to the worktree, hence common=False)
        cherryPickHead = Path(self.repo.in_gitdir("CHERRY_PICK_HEAD", common=False))
        cherryPickHead.write_text(str(oid) + "\n")

        # Refresh libgit2 index for conflict analysis
        self.repo.refresh_index()

        anyConflicts = exitCode != 0
        dud = not anyConflicts and not self.repo.any_staged_changes

        assert self.repo.state() == RepositoryState.CHERRYPICK

        # If cherrypicking didn't do anything, don't let the CHERRYPICK state linger.
        # (Otherwise, the state will be cleared when we commit)
        if dud:
            self.repo.state_cleanup()

        # Back to UI thread
        yield from self.flowEnterUiThread()

        if dud:
            info = _("There’s nothing to cherry-pick from {0} "
                     "that the current branch doesn’t already have.", bquo(shortHash(oid)))
            raise AbortTask(info, "information")

        self.repoModel.prefs.draftCommitMessage = self.repo.message_without_conflict_comments
        self.repoModel.prefs.draftCommitSignature = commit.author
        self.repoModel.prefs.draftCommitSignatureOverride = SignatureOverride.Author
        self.repoModel.prefs.setDirty()

        self.epilog.jumpTo = NavLocator.inWorkdir()



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations

import bisect
import collections
from collections.abc import Sequence

from gitfourchette.graph.graph import Graph, Frame, Oid, Arc, ChainHandle, BatchRow, BATCHROW_UNDEF, ArcJunction


class GraphWeaver(Frame):
    freeLanes: list[int]
    parentLookup: collections.defaultdict[Oid, list[Arc]]  # list: all lanes
    peakArcCount: int
    batchNo: int

    @staticmethod
    def newGraph() -> tuple[Graph, GraphWeaver]:
        graph = Graph()
        assert graph.isEmpty(), "cannot regenerate an existing graph!"
        weaver = GraphWeaver(graph.startArc)
        graph.ownBatches.append(weaver.batchNo)
        return graph, weaver

    def __init__(self, startArcSentinel: Arc):
        super().__init__(row=BATCHROW_UNDEF, commit=None,
                         solvedArcs=[], openArcs=[], lastArc=startArcSentinel)
        self.freeLanes = []
        self.parentLookup = collections.defaultdict(list)
        self.peakArcCount = 0
        self.batchNo = BatchRow.BatchManager.reserveNewBatch()

    def newCommit(self, me: Oid, myParents: Sequence[Oid]):
        """Create arcs for a new commit row."""

        row = BatchRow(b=self.batchNo, y=self.row.y + 1)
        self.row = row
        self.commit = me

        hasParents = len(myParents) > 0

        # Resolve arcs that my child commits have opened higher up in the graph,
        # waiting for me to appear in the commit sequence so I can close them.
        myOpenArcs = self.parentLookup.get(me)
        myHomeLane = -1
        handOffHomeLane = False
        if not myOpenArcs:
            # Nobody was looking for me, so I'm the tip of a new branch
            myHomeChain = ChainHandle(row, BATCHROW_UNDEF)
        else:
            myMainOpenArc = min(myOpenArcs, key=lambda a: a.lane)
            myHomeLane = myMainOpenArc.lane
            myHomeChain = myMainOpenArc.chain
            assert myHomeChain.isValid()
            for arc in myOpenArcs:
                # Close off open arcs
                assert arc.closedBy == me
                assert arc.closedAt == BATCHROW_UNDEF
                assert arc.chain.bottomRow == BATCHROW_UNDEF
                arc.closedAt = row
                self.solvedArcs[arc.lane] = arc
                self.openArcs[arc.lane] = None  # Free up the lane below
                if hasParents and arc.lane == myHomeLane:
                    handOffHomeLane = True
                else:
                    bisect.insort(self.freeLanes, arc.lane)
                    # Close off chain
                    arc.chain.bottomRow = row
            del self.parentLookup[me]

            """
            # Compact null arcs at right of graph
            if not allocLanesInGaps:
                for _ in range(len(self.openArcs)-1, myHomeLane, -1):
                    if self.openArcs[-1] is not None:
                        break
                    self.openArcs.pop()
                    self.solvedArcs.pop()
            """

        sawFirstParent = False
        for parent in myParents:
            # See if there's already an arc that is looking for any of my parents BEYOND PARENT ZERO.
            # If so, make a junction on that arc.
            if sawFirstParent:
                arcsOfParent = self.parentLookup.get(parent)
                if arcsOfParent:
                    arc = min(arcsOfParent, key=lambda a: a.lane)
                    assert arc.closedBy == parent
                    assert arc.closedAt == BATCHROW_UNDEF
                    arc.junctions.append(ArcJunction(joinedAt=row, joinedBy=me))
                    continue

            # We didn't make a junction, so open up a new arc on a free lane.
            # Get a free lane.
            if not sawFirstParent and myHomeLane >= 0:
                assert handOffHomeLane
                freeLane = myHomeLane
                handOffHomeLane = False
            elif not self.freeLanes:
                # Allocate new lane on the right
                freeLane = len(self.openArcs)
                self.openArcs.append(None)  # Reserve the lane
                self.solvedArcs.append(None)  # Reserve the lane
            else:
                # Pick leftmost free lane
                freeLane = self.freeLanes.pop(0)

            if not sawFirstParent:
                # Hand off my chain
                parentChain = myHomeChain
            else:
                # Branch out new chain downward
                parentChain = ChainHandle(row, BATCHROW_UNDEF)
            assert parentChain.bottomRow == BATCHROW_UNDEF

            # Make arc from this commit to its parent
            newArc = Arc(lane=freeLane, chain=parentChain,
                         openedAt=row, closedAt=BATCHROW_UNDEF,
                         openedBy=me, closedBy=parent, junctions=[])
            self.openArcs[freeLane] = newArc
            self.parentLookup[parent].append(newArc)
            self.lastArc.nextArc = newArc
            self.lastArc = newArc

            sawFirstParent = True

        # Parentless commit: we'll insert a bogus arc so playback doesn't need the commit sequence.
        # This is just to keep the big linked list of arcs flowing.
        # Do NOT put it in solved or open arcs! This would interfere with detection of arcs that the commit
        # actually closes or opens.
        if not sawFirstParent:
            if myHomeLane < 0:
                # Edge case: the commit is BOTH childless AND parentless.
                # Put the commit in a free lane, but do NOT reserve the lane.
                if not self.freeLanes:
                    myHomeLane = len(self.openArcs)
                else:
                    myHomeLane = self.freeLanes[0]  # PEEK only! do not pop!

                # Close off home chain
                myHomeChain.bottomRow = row

            assert myHomeChain.bottomRow.isValid(), "parentless commit's home chain shouldn't be dangling at bottom"

            newArc = Arc(lane=myHomeLane, chain=myHomeChain,
                         openedAt=row, closedAt=row,
                         openedBy=me, closedBy=me, junctions=[])
            self.lastArc.nextArc = newArc
            self.lastArc = newArc
            assert newArc.isParentlessCommitBogusArc()

        assert not handOffHomeLane
        assert myHomeChain.topRow.isValid()
        assert len(self.openArcs) == len(self.solvedArcs)

        # Keep track of peak arc count for statistics
        self.peakArcCount = max(self.peakArcCount, len(self.openArcs))

    def isDangling(self):
        return len(self.parentLookup) > 0



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from collections.abc import Callable

from gitfourchette import colors
from gitfourchette.localization import *
from gitfourchette.nav import NavLocator, NavContext
from gitfourchette.qt import *
from gitfourchette.tasks import *
from gitfourchette.toolbox import *

PERMANENT_PROPERTY = "permanent"


class ContextHeader(QFrame):
    def __init__(self, parent):
        super().__init__(parent)
        self.setObjectName("ContextHeader")
        self.setSizePolicy(QSizePolicy.Policy.Ignored, QSizePolicy.Policy.Minimum)

        self.locator = NavLocator()
        self.buttons = []

        self.mainLabel = QLabel(self)
        self.spacerItem = QSpacerItem(0, 0, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        layout = QHBoxLayout(self)
        layout.setContentsMargins(2, 0, 0, 0)
        layout.addWidget(self.mainLabel)
        layout.addSpacerItem(self.spacerItem)

        self.maximizeButton = self.addButton(_("Maximize"), permanent=True)
        self.maximizeButton.setIcon(stockIcon("maximize"))
        self.maximizeButton.setToolTip(_("Maximize the diff area and hide the commit graph"))
        self.maximizeButton.setCheckable(True)

    def addButton(
            self,
            text: str,
            callback: Callable | None = None,
            permanent: bool = False,
            stickToLabel: bool = False,
    ) -> QToolButton:
        button = QToolButton(self)
        button.setText(text)
        button.setProperty(PERMANENT_PROPERTY, "true" if permanent else "")
        button.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        button.setAutoRaise(True)
        button.setMaximumHeight(20)
        self.buttons.append(button)

        if callback is not None:
            button.clicked.connect(callback)

        layout = self.layout()
        assert isinstance(layout, QBoxLayout)
        insertionIndex = layout.indexOf(self.spacerItem)
        if not stickToLabel:
            insertionIndex += 1
        layout.insertWidget(insertionIndex, button)

        return button

    def clearButtons(self):
        for i in range(len(self.buttons) - 1, -1, -1):
            button = self.buttons[i]
            if not button.property(PERMANENT_PROPERTY):
                button.hide()
                button.deleteLater()
                del self.buttons[i]

    @DisableWidgetUpdatesContext.methodDecorator
    def setContext(
            self,
            locator: NavLocator,
            commitMessage: str = "",
            isStash=False,
    ):
        self.clearButtons()

        self.locator = locator

        introStyle = "font-weight: 500;"
        mainText = " "

        diffAB = locator.commitDiffAB()
        if diffAB:
            mainText = (
                "<span style='{introStyle}'>{intro}{colon}</span> "
                "<span style='color: {red}; font-weight: bold;'>A: </span> {a} \u2192 "
                "<span style='color: {blue}; font-weight: bold;'>B: </span> {b}"
            ).format(introStyle=introStyle, intro=_("Comparing commits"), colon=_(":"),
                     red=colors.red.name(), blue=colors.blue.name(),
                     a=shortHash(diffAB[0]), b=shortHash(diffAB[1]))

            swapButton = self.addButton(
                _("Swap A/B"),
                lambda: Jump.invoke(self, locator.swapABCommits()),
                stickToLabel=True)

            swapButton.setToolTip(_("Swap the “old” and “new” sides in the commit comparison"))

        elif locator.selectedCommits:
            pass

        elif locator.context == NavContext.COMMITTED:
            kind = _p("noun", "Stash") if isStash else _p("noun", "Commit")
            summary, _continued = messageSummary(commitMessage)
            mainText = "<span style='{introStyle}'>{intro} {hash}{colon}</span> {summary}".format(
                introStyle=introStyle, intro=kind, hash=shortHash(locator.commit), colon=_(":"), summary=escape(summary))

            infoButton = self.addButton(_("Info"), lambda: GetCommitInfo.invoke(self, self.locator.commit))
            infoButton.setToolTip(_("Show details about this commit") if not isStash
                                  else _("Show details about this stash"))

            if isStash:
                dropButton = self.addButton(_("Delete"), lambda: DropStash.invoke(self, locator.commit))
                dropButton.setToolTip(_("Delete this stash"))

                applyButton = self.addButton(_("Apply"), lambda: ApplyStash.invoke(self, locator.commit))
                applyButton.setToolTip(_("Apply this stash"))

        elif locator.context.isWorkdir():
            mainText = _("Working Directory")

        self.mainLabel.setText(mainText)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.forms.brandeddialog import convertToBrandedDialog
from gitfourchette.forms.ui_newtagdialog import Ui_NewTagDialog
from gitfourchette.localization import *
from gitfourchette.qt import *
from gitfourchette.toolbox import *


def populateRemoteComboBox(comboBox: QComboBox, remotes: list[str]):
    assert 0 == comboBox.count()
    if not remotes:
        comboBox.addItem(_("No Remotes"))
    else:
        comboBox.addItem(_("All Remotes"), userData="*")
        comboBox.insertSeparator(1)
        for remote in remotes:
            comboBox.addItem(remote, userData=remote)


class NewTagDialog(QDialog):
    def __init__(
            self,
            target: str,
            targetSubtitle: str,
            reservedNames: list[str],
            remotes: list[str],
            parent=None):

        super().__init__(parent)

        self.ui = Ui_NewTagDialog()
        self.ui.setupUi(self)

        self.reservedNames = reservedNames

        okButton = self.ui.buttonBox.button(QDialogButtonBox.StandardButton.Ok)
        okButton.setIcon(stockIcon("git-tag"))
        okCaptions = [_("&Create"), _("&Create && Push")]
        self.ui.pushCheckBox.toggled.connect(lambda push: okButton.setText(okCaptions[push]))

        populateRemoteComboBox(self.ui.remoteComboBox, remotes)

        # Enable/disable OK button depending on input
        validator = ValidatorMultiplexer(self)
        validator.setGatedWidgets(okButton)
        validator.connectInput(self.ui.nameEdit, self.validateOK)
        validator.run(silenceEmptyWarnings=True)

        # Enable/disable 'Replace' checkbox depending on input
        validatorForce = ValidatorMultiplexer(self)
        validatorForce.setGatedWidgets(self.ui.forceCheckBox)
        validatorForce.connectInput(self.ui.nameEdit, self.validateForce, showError=False)
        validatorForce.run(silenceEmptyWarnings=True)

        self.ui.forceCheckBox.toggled.connect(validator.run)

        # Prime enabled state
        self.ui.pushCheckBox.click()
        self.ui.pushCheckBox.click()
        if not remotes:
            self.ui.pushCheckBox.setChecked(False)
            self.ui.pushCheckBox.setEnabled(False)

        convertToBrandedDialog(
            self,
            _("New tag on commit {0}", tquo(target)),
            tquo(targetSubtitle))

        packDialog(self, lockHeight=True)

    def validateOK(self, name: str) -> str:
        nameTaken = _("This name is already taken by another tag.")

        reservedNames = self.reservedNames
        if self.ui.forceCheckBox.isEnabled() and self.ui.forceCheckBox.isChecked():
            reservedNames = []

        return nameValidationMessage(name, reservedNames, nameTaken)

    def validateForce(self, name: str) -> str:
        # The 'replace' checkbox will be DISABLED if the validation string
        # evaluates to False (as a bool)
        taken = name.lower() in (n.lower() for n in self.reservedNames)
        return "" if taken else "DISABLE_FORCE_CHECKBOX"



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations

import logging
import os
import re
from pathlib import Path

from gitfourchette import settings
from gitfourchette.codeview.codeview import CodeView
from gitfourchette.diffview.diffdocument import DiffDocument, LineData
from gitfourchette.diffview.diffgutter import DiffGutter
from gitfourchette.diffview.diffhighlighter import DiffHighlighter
from gitfourchette.gitdriver import GitDelta
from gitfourchette.globalshortcuts import GlobalShortcuts
from gitfourchette.localization import *
from gitfourchette.nav import NavContext, NavLocator
from gitfourchette.porcelain import *
from gitfourchette.qt import *
from gitfourchette.subpatch import extractSubpatch
from gitfourchette.tasks import ApplyPatch, ApplyPatchData
from gitfourchette.tasks.blametasks import OpenBlameToLine
from gitfourchette.toolbox import *

logger = logging.getLogger(__name__)
_emptyDelta = GitDelta()


class DiffView(CodeView):
    contextualHelp = Signal(str)
    selectionActionable = Signal(bool)
    visibilityChanged = Signal(bool)

    lineData: list[LineData]
    currentLocator: NavLocator
    currentDelta: GitDelta
    currentDiffDocument: DiffDocument | None
    repo: Repo | None

    def __init__(self, parent=None):
        super().__init__(gutterClass=DiffGutter, highlighterClass=DiffHighlighter, parent=parent)

        self.lineData = []
        self.currentLocator = NavLocator.Empty
        self.currentDelta = _emptyDelta
        self.currentDiffDocument = None
        self.repo = None

        # Emit contextual help with non-empty selection
        self.cursorPositionChanged.connect(self.emitSelectionHelp)
        self.selectionChanged.connect(self.emitSelectionHelp)

        self._initRubberBandButtons()

        self.gutter.lineClicked.connect(self.selectWholeLineAt)
        self.gutter.lineShiftClicked.connect(self.selectWholeLinesTo)
        self.gutter.lineDoubleClicked.connect(self.selectClumpOfLinesAt)
        self.gutter.selectionMiddleClicked.connect(self.onMiddleClick)

    def _initRubberBandButtons(self):
        self.stageButton = QToolButton()
        self.stageButton.setText(_("Stage Selection"))
        self.stageButton.setIcon(stockIcon("git-stage-lines"))
        self.stageButton.setToolTip(appendShortcutToToolTipText(_("Stage selected lines"), GlobalShortcuts.stageHotkeys[0]))
        self.stageButton.clicked.connect(self.stageSelection)

        self.unstageButton = QToolButton()
        self.unstageButton.setText(_("Unstage Selection"))
        self.unstageButton.setIcon(stockIcon("git-unstage-lines"))
        self.unstageButton.clicked.connect(self.unstageSelection)
        self.unstageButton.setToolTip(appendShortcutToToolTipText(_("Unstage selected lines"), GlobalShortcuts.discardHotkeys[0]))

        self.discardButton = QToolButton()
        self.discardButton.setText(_("Discard"))
        self.discardButton.setIcon(stockIcon("git-discard-lines"))
        self.discardButton.clicked.connect(self.discardSelection)
        self.discardButton.setToolTip(appendShortcutToToolTipText(_("Discard selected lines"), GlobalShortcuts.discardHotkeys[0]))

        for button in self.stageButton, self.discardButton, self.unstageButton:
            button.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
            button.setCursor(Qt.CursorShape.PointingHandCursor)
            button.setFocusPolicy(Qt.FocusPolicy.NoFocus)
            self.rubberBandButtonGroup.layout().addWidget(button)

        makeWidgetShortcut(self, self.onStageShortcut, *GlobalShortcuts.stageHotkeys)
        makeWidgetShortcut(self, self.onDiscardShortcut, *GlobalShortcuts.discardHotkeys)

    # ---------------------------------------------
    # Callbacks for Qt events/shortcuts

    def onStageShortcut(self):
        navContext = self.currentLocator.context
        if navContext == NavContext.UNSTAGED:
            self.stageSelection()
        else:
            QApplication.beep()

    def onDiscardShortcut(self):
        navContext = self.currentLocator.context
        if navContext == NavContext.STAGED:
            self.unstageSelection()
        elif navContext == NavContext.UNSTAGED:
            self.discardSelection()
        else:
            QApplication.beep()

    def mouseReleaseEvent(self, event: QMouseEvent):
        super().mouseReleaseEvent(event)
        if event.button() == Qt.MouseButton.MiddleButton:
            self.onMiddleClick()

    # ---------------------------------------------
    # Document replacement

    def clear(self):  # override
        # Clear info about the current patch - necessary for document reuse detection to be correct when the user
        # clears the selection in a FileList and then reselects the last-displayed document.
        self.currentDelta = _emptyDelta

        # Clear the actual contents
        super().clear()

    @benchmark
    def replaceDocument(self, repo: Repo, delta: GitDelta, locator: NavLocator, newDoc: DiffDocument):
        assert newDoc.document is not None
        assert isinstance(self.highlighter, DiffHighlighter)
        assert isinstance(self.gutter, DiffGutter)

        # No-op if reloading same document
        if newDoc is self.currentDiffDocument:
            return

        oldDocument = self.document()
        if oldDocument:
            oldDocument.deleteLater()  # avoid leaking memory/objects, even though we do set QTextDocument's parent to this QTextEdit

        self.repo = repo
        self.currentDelta = delta
        self.currentLocator = locator
        self.currentDiffDocument = newDoc
        self.lineData = newDoc.lineData

        newDoc.document.setParent(self)
        self.setDocument(newDoc.document)
        self.highlighter.setDiffDocument(newDoc)
        self.searchBar.reevaluateSearchTerm()

        # now reset defaults that are lost when changing documents
        self.refreshPrefs(changeColorScheme=False)

        self.gutter.maxLine = newDoc.maxLine
        self.syncViewportMarginsWithGutter()

        buttonMask = 0
        if locator.context == NavContext.UNSTAGED:
            buttonMask = PatchPurpose.Stage | PatchPurpose.Discard
        elif locator.context == NavContext.STAGED:
            buttonMask = PatchPurpose.Unstage
        self.stageButton.setVisible(bool(buttonMask & PatchPurpose.Stage))
        self.discardButton.setVisible(bool(buttonMask & PatchPurpose.Discard))
        self.unstageButton.setVisible(bool(buttonMask & PatchPurpose.Unstage))

        # Now restore cursor/scrollbar positions
        self.restorePosition(locator)

    # ---------------------------------------------
    # Context menu

    def contextMenuActions(self, clickedCursor: QTextCursor) -> list[ActionDef]:
        # If we have a document, we should have a patch
        assert self.currentDelta != _emptyDelta

        cursor: QTextCursor = self.textCursor()
        hasSelection = cursor.hasSelection()

        # Find hunk at click position
        block = self.document().findBlock(clickedCursor.position())
        blockNumber = block.blockNumber()

        try:
            lineData = self.lineData[blockNumber]
        except IndexError:
            lineData = None
            shortHunkHeader = "???"
        else:
            clickedHunkID = lineData.hunkPos.hunkID
            hunkFirstLine, _hunkLastLine = LineData.getHunkExtents(self.lineData, clickedHunkID)
            headerMatch = re.match(r"@@ ([^@]+) @@.*", self.lineData[hunkFirstLine].text)
            shortHunkHeader = headerMatch.group(1) if headerMatch else f"#{clickedHunkID}"

        actions = []

        navContext = self.currentLocator.context

        if navContext == NavContext.COMMITTED:
            if hasSelection:
                actions = [
                    ActionDef(_("Export Lines as Patch…"), self.exportSelection),
                    ActionDef(_("Revert Lines…"), self.revertSelection),
                ]
            else:
                actions = [
                    ActionDef(_("Export Hunk {0} as Patch…", shortHunkHeader), lambda: self.exportHunk(clickedHunkID)),
                    ActionDef(_("Revert Hunk…"), lambda: self.revertHunk(clickedHunkID)),
                ]

        elif navContext == NavContext.UNSTAGED:
            if hasSelection:
                actions = [
                    ActionDef(
                        _("Stage Lines"),
                        self.stageSelection,
                        "git-stage-lines",
                        shortcuts=GlobalShortcuts.stageHotkeys[0],
                    ),
                    ActionDef(
                        _("Discard Lines…"),
                        self.discardSelection,
                        "git-discard-lines",
                        shortcuts=GlobalShortcuts.discardHotkeys[0],
                    ),
                    ActionDef(
                        _("Export Lines as Patch…"),
                        self.exportSelection
                    ),
                ]
            else:
                actions = [
                    ActionDef(
                        _("Stage Hunk {0}", shortHunkHeader),
                        lambda: self.stageHunk(clickedHunkID),
                        "git-stage-lines",
                    ),
                    ActionDef(
                        _("Discard Hunk…"),
                        lambda: self.discardHunk(clickedHunkID),
                        "git-discard-lines",
                    ),
                    ActionDef(_("Export Hunk as Patch…"), lambda: self.exportHunk(clickedHunkID)),
                ]

        elif navContext == NavContext.STAGED:
            if hasSelection:
                actions = [
                    ActionDef(
                        _("Unstage Lines"),
                        self.unstageSelection,
                        "git-unstage-lines",
                        shortcuts=GlobalShortcuts.discardHotkeys[0],
                    ),
                    ActionDef(
                        _("Export Lines as Patch…"),
                        self.exportSelection,
                    ),
                ]
            else:
                actions = [
                    ActionDef(
                        _("Unstage Hunk {0}", shortHunkHeader),
                        lambda: self.unstageHunk(clickedHunkID),
                        "git-unstage-lines",
                    ),
                    ActionDef(
                        _("Export Hunk as Patch…"),
                        lambda: self.exportHunk(clickedHunkID),
                    ),
                ]

        blamePreview = ""
        allowBlameLine = (
                lineData is not None
                and bool(lineData.origin)
                and not (self.currentDelta.source.isWorkdir() and self.currentDelta.status.isAddedOrUntracked)
        )
        if allowBlameLine:
            assert not lineData.hunkPos.isHunkHeaderLine(), "line had truthy origin; not expecting hunk header"
            blamePreview = lineData.text.strip()
            if blamePreview:
                blamePreview = elide(blamePreview, Qt.TextElideMode.ElideRight, 15)
                blamePreview = lquo(blamePreview)
            else:
                blamePreview += f"-{lineData.oldLineNo}" if lineData.origin == "-" else f"+{lineData.newLineNo}"

        actions += [
            ActionDef(
                _("Blame Line") + " " + blamePreview,
                lambda: self.blameLine(lineData),
                enabled=allowBlameLine,
            ),
        ]

        return actions

    # ---------------------------------------------
    # Patch

    def findHunkIDAt(self, cursorPosition: int):
        block = self.document().findBlock(cursorPosition)
        blockNumber = block.blockNumber()
        try:
            return self.lineData[blockNumber].hunkPos.hunkID
        except IndexError:
            return -1

    def isSelectionActionable(self):
        start, end = self.getSelectedLineExtents()
        numAdds = 0
        numDels = 0
        for i in range(start, end+1):
            try:
                ld = self.lineData[i]
            except IndexError:
                assert (numAdds, numDels) == (0, 0)
                break
            if ld.origin == "+":
                numAdds += 1
            elif ld.origin == "-":
                numDels += 1
        return numAdds, numDels

    def extractSelection(self, reverse=False) -> str:
        i, j = self.getSelectedLineExtents()
        return extractSubpatch(self.currentDelta, self.lineData, i, j, reverse)

    def extractHunk(self, hunkID: int, reverse=False) -> str:
        i, j = LineData.getHunkExtents(self.lineData, hunkID)
        return extractSubpatch(self.currentDelta, self.lineData, i, j, reverse)

    def exportPatch(self, patchData: str):
        if not patchData:
            QApplication.beep()
            return

        def dump(path: str):
            Path(path).write_text(patchData, encoding="utf-8", newline="\n")

        name = os.path.basename(self.currentLocator.path) + "[partial].patch"
        qfd = PersistentFileDialog.saveFile(self, "SaveFile", _("Export selected lines"), name)
        qfd.fileSelected.connect(dump)
        qfd.show()

    def fireApplyLines(self, purpose: PatchPurpose):
        purpose |= PatchPurpose.Lines
        reverse = not (purpose & PatchPurpose.Stage)
        patchData = self.extractSelection(reverse)
        ApplyPatch.invoke(self, self.currentDelta, patchData, purpose)

    def fireApplyHunk(self, hunkID: int, purpose: PatchPurpose):
        purpose |= PatchPurpose.Hunk
        reverse = not (purpose & PatchPurpose.Stage)
        patchData = self.extractHunk(hunkID, reverse)
        ApplyPatch.invoke(self, self.currentDelta, patchData, purpose)

    def onMiddleClick(self):
        if not settings.prefs.middleClickStageLines:
            return

        navContext = self.currentLocator.context
        if navContext == NavContext.UNSTAGED:
            self.stageSelection()
        elif navContext == NavContext.STAGED:
            self.unstageSelection()
        else:
            QApplication.beep()

    def blameLine(self, ld: LineData):
        OpenBlameToLine.invoke(self, self.currentDelta, ld)

    def stageSelection(self):
        self.fireApplyLines(PatchPurpose.Stage)

    def unstageSelection(self):
        self.fireApplyLines(PatchPurpose.Unstage)

    def discardSelection(self):
        self.fireApplyLines(PatchPurpose.Discard)

    def exportSelection(self):
        patchData = self.extractSelection()
        self.exportPatch(patchData)

    def revertSelection(self):
        """ Revert selected lines (typically in an older commit). """

        # Prepare a pre-reverted patch for best results
        patchData = self.extractSelection(reverse=True)

        # Put a cap on how much context git will look at - this makes the patch
        # more likely to apply if the files have diverged.
        # (extractSelection keeps the entire hunk as context, some of which
        # may be obsolete in an older commit)
        ApplyPatchData.invoke(self, patchData, context=3,
                              title=_("Revert selected lines"),
                              question=_("Do you want to revert the selected lines?"))

    def stageHunk(self, hunkID: int):
        self.fireApplyHunk(hunkID, PatchPurpose.Stage)

    def unstageHunk(self, hunkID: int):
        self.fireApplyHunk(hunkID, PatchPurpose.Unstage)

    def discardHunk(self, hunkID: int):
        self.fireApplyHunk(hunkID, PatchPurpose.Discard)

    def exportHunk(self, hunkID: int):
        patchData = self.extractHunk(hunkID)
        self.exportPatch(patchData)

    def revertHunk(self, hunkID: int):
        """ Revert a hunk (typically in an older commit). """

        # Prepare a pre-reverted patch for best results
        patchData = self.extractHunk(hunkID, reverse=True)

        # Put a cap on how much context git will look at - this makes the patch
        # more likely to apply if the files have diverged.
        ApplyPatchData.invoke(self, patchData, context=3,
                              title=_("Revert hunk"),
                              question=_("Do you want to revert this hunk?"))

    # ---------------------------------------------
    # Rubberband

    def updateRubberBand(self):
        textCursor: QTextCursor = self.textCursor()
        start = textCursor.selectionStart()
        end = textCursor.selectionEnd()
        assert start <= end

        startLine, endLine = self.getSelectedLineExtents()
        numAdds, numDels = self.isSelectionActionable()
        actionable = numAdds + numDels > 0

        if startLine < 0 or endLine < 0 or (not actionable and start == end):
            self.rubberBand.hide()
            self.rubberBandButtonGroup.hide()
            return

        start = self.lineData[startLine].cursorStart
        end = self.lineData[endLine].cursorEnd

        textCursor.setPosition(start)
        top = self.cursorRect(textCursor).top()

        textCursor.setPosition(end)
        bottom = self.cursorRect(textCursor).bottom()

        viewportWidth = self.viewport().width()
        viewportHeight = self.viewport().height()

        pad = 4
        self.rubberBand.setGeometry(0, top-pad, viewportWidth, bottom-top+1+pad*2)
        self.rubberBand.show()

        # Move rubberBandButton to edge of rubberBand
        if not self.currentLocator.context.isWorkdir():
            # No rubberBandButton in this context
            assert self.rubberBandButtonGroup.isHidden()
        elif top >= viewportHeight-4 or bottom < 4:
            # Scrolled past rubberband, no point in showing button
            self.rubberBandButtonGroup.hide()
        else:
            # Show button before moving so that width/height are correct
            self.rubberBandButtonGroup.show()
            self.rubberBandButtonGroup.ensurePolished()
            rbbWidth = self.rubberBandButtonGroup.width()
            rbbHeight = self.rubberBandButtonGroup.height()

            # Place button above rubberband
            rbbTop = top - rbbHeight
            rbbTop = max(rbbTop, 0)  # keep it visible
            self.rubberBandButtonGroup.move(viewportWidth - rbbWidth, rbbTop)
            self.rubberBandButtonGroup.setEnabled(actionable)

    # ---------------------------------------------
    # Cursor/selection

    def selectClumpOfLinesAt(self, clickPoint: QPoint | None = None, textCursorPosition: int = -1):
        assert (textCursorPosition >= 0) ^ (clickPoint is not None)
        if textCursorPosition < 0:
            assert clickPoint is not None
            textCursorPosition = self.getStartOfLineAt(clickPoint)

        ldList = self.lineData
        i = self.document().findBlock(textCursorPosition).blockNumber()
        ld = ldList[i]

        if ld.hunkPos.hunkLineNum < 0:
            # Hunk header line, select whole hunk
            start = i
            end = i
            while end < len(ldList)-1 and ldList[end+1].hunkPos.hunkID == ld.hunkPos.hunkID:
                end += 1
        elif ld.clumpID < 0:
            # Context line
            QApplication.beep()
            return
        else:
            # Get clump boundaries
            start = i
            end = i
            while start > 0 and ldList[start-1].clumpID == ld.clumpID:
                start -= 1
            while end < len(ldList)-1 and ldList[end+1].clumpID == ld.clumpID:
                end += 1

        startPosition = ldList[start].cursorStart
        endPosition = min(self.getMaxPosition(), ldList[end].cursorEnd)

        cursor: QTextCursor = self.textCursor()
        cursor.setPosition(startPosition, QTextCursor.MoveMode.MoveAnchor)
        cursor.setPosition(endPosition, QTextCursor.MoveMode.KeepAnchor)
        self.replaceCursor(cursor)

    # ---------------------------------------------
    # Selection help

    def emitSelectionHelp(self):
        if self.currentLocator.context in [NavContext.COMMITTED, NavContext.EMPTY]:
            return

        numAdds, numDels = self.isSelectionActionable()

        if numAdds + numDels == 0:
            self.contextualHelp.emit("")
            self.selectionActionable.emit(False)
            return

        stageKey = QKeySequence(GlobalShortcuts.stageHotkeys[0]).toString(QKeySequence.SequenceFormat.NativeText)
        discardKey = QKeySequence(GlobalShortcuts.discardHotkeys[0]).toString(QKeySequence.SequenceFormat.NativeText)
        unstageKey = discardKey
        if settings.prefs.middleClickStageLines:
            stageKey += " " + _("or Middle-Click")
            unstageKey += " " + _("or Middle-Click")

        if numAdds and numDels:
            nl = f"+{numAdds} -{numDels}"
        else:
            nl = f"+{numAdds}" if numAdds else f"-{numDels}"
        if self.currentLocator.context == NavContext.UNSTAGED:
            help = _n("Hit {sk} to stage {nl} line. Hit {dk} to discard it.",
                      "Hit {sk} to stage {nl} lines. Hit {dk} to discard them.",
                      n=numAdds+numDels, nl=nl, sk=stageKey, dk=discardKey)
        elif self.currentLocator.context == NavContext.STAGED:
            help = _n("Hit {uk} to unstage {nl} line.",
                      "Hit {uk} to unstage {nl} lines.",
                      n=numAdds+numDels, nl=nl, uk=unstageKey)

        self.contextualHelp.emit(help)
        self.selectionActionable.emit(True)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.forms.brandeddialog import convertToBrandedDialog
from gitfourchette.forms.ui_remotedialog import Ui_RemoteDialog
from gitfourchette.localization import *
from gitfourchette.qt import *
from gitfourchette.toolbox import *


class RemoteDialog(QDialog):
    def __init__(
            self,
            editExistingRemote: bool,
            name: str,
            url: str,
            existingRemotes: list[str],
            skipFetchAll: bool,
            parent: QWidget
    ):
        super().__init__(parent)

        self.existingRemotes = existingRemotes

        self.ui = Ui_RemoteDialog()
        self.ui.setupUi(self)
        okButton = self.ui.buttonBox.button(QDialogButtonBox.StandardButton.Ok)

        self.ui.urlEdit.setText(url)
        self.ui.nameEdit.setText(name)
        self.ui.skipFetchAllCheckBox.setChecked(skipFetchAll)
        formatWidgetTooltip(self.ui.skipFetchAllCheckBox, cmd=hquo("git fetch --all"), forceWrap=True)

        # Set up input validator
        nameTaken = _("This name is already taken by another remote.")
        cannotBeEmpty = _("Cannot be empty.")
        validator = ValidatorMultiplexer(self)
        validator.setGatedWidgets(okButton)
        validator.connectInput(self.ui.nameEdit, lambda s: nameValidationMessage(s, existingRemotes, nameTaken))
        validator.connectInput(self.ui.urlEdit, lambda s: cannotBeEmpty if not s.strip() else "")

        # Different behavior if editing existing remote or creating a new one
        if editExistingRemote:
            # Edit existing remote
            title = _("Edit remote {0}", hquoe(name))
            self.setWindowTitle(_("Edit remote"))
            okButton.setText(_("Save"))
            self.ui.fetchAfterAddCheckBox.setVisible(False)
            # Don't touch name automatically on existing remotes
            self.allowAutoFillName = False
        else:
            # Create new remote
            title = _("Add remote")
            self.setWindowTitle(_("Add remote"))
            okButton.setText(_("Add"))
            self.ui.fetchAfterAddCheckBox.setVisible(True)
            # Autofill name on new remotes
            self.allowAutoFillName = True
            self.ui.urlEdit.textChanged.connect(self.onUrlChangedAutoFillName)
            self.ui.nameEdit.textChanged.connect(self.onNameChanged)
            # Automatically paste URL in clipboard
            if not url:
                url = guessRemoteUrlFromText(QApplication.clipboard().text())
                self.ui.urlEdit.setText(url)
                self.ui.urlEdit.setFocus()

        # Connect protocol button to URL editor
        self.ui.protocolButton.connectTo(self.ui.urlEdit)

        convertToBrandedDialog(self, title)
        packDialog(self, lockHeight=True)

        # Run input callback
        validator.run(silenceEmptyWarnings=True)

    def onNameChanged(self, name: str):
        # Allow autofilling name again if user has erased it
        self.allowAutoFillName = name == ""

    def onUrlChangedAutoFillName(self, url: str):
        if not self.allowAutoFillName:
            return

        url = url.strip()
        host, _path = splitRemoteUrl(url)

        # Sanitize host
        for c in " ?/\\*~<>|:":
            host = host.replace(c, "_")

        # Clean up common host names (git.(...).org)
        host = host.removeprefix("www.")
        host = host.removeprefix("git.")
        for tld in (".com", ".org", ".net"):
            if host.endswith(tld):
                host = host.removesuffix(tld)
                break

        host = withUniqueSuffix(host, self.existingRemotes)
        self.ui.nameEdit.setText(host)
        self.allowAutoFillName = True  # re-enable this since we got this far



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import logging
import os
import shutil
from pathlib import Path
from collections.abc import Callable

from gitfourchette import trtables
from gitfourchette import settings
from gitfourchette.codeview.codewindow import CodeWindow
from gitfourchette.exttools.mergedriver import MergeDriver
from gitfourchette.exttools.toolprocess import ToolProcess
from gitfourchette.gitdriver import *
from gitfourchette.localization import *
from gitfourchette.nav import NavLocator
from gitfourchette.porcelain import *
from gitfourchette.qt import *
from gitfourchette.tasks.repotask import AbortTask, RepoTask, TaskEffects
from gitfourchette.toolbox import *
from gitfourchette.trash import Trash

logger = logging.getLogger(__name__)


class _BaseStagingTask(RepoTask):
    def canKill(self, task: RepoTask):
        # Jump/Refresh tasks shouldn't prevent a staging task from starting
        # when the user holds down RETURN/DELETE in a FileListView
        # to stage/unstage a series of files.
        from gitfourchette import tasks
        return isinstance(task, tasks.Jump | tasks.RefreshRepo)

    def denyConflicts(self, deltas: list[GitDelta], purpose: PatchPurpose):
        # Filter on conflicts
        conflictPaths = [d.new.path for d in deltas if d.conflict is not None]

        # Bail if no conflicts
        if not conflictPaths:
            return

        numPatches = len(deltas)
        numConflicts = len(conflictPaths)

        if numPatches == numConflicts:
            intro = _n("You have selected an unresolved merge conflict.",
                       "You have selected {n} unresolved merge conflicts.", numConflicts)
        else:
            intro = _n("There is an unresolved merge conflict among your selection.",
                       "There are {n} unresolved merge conflicts among your selection.", numConflicts)

        if purpose == PatchPurpose.Stage:
            please = _np("please fix (the merge conflicts)", "Please fix it before staging:", "Please fix them before staging:", numConflicts)
        else:
            please = _np("please fix (the merge conflicts)", "Please fix it before discarding:", "Please fix them before discarding:", numConflicts)

        message = paragraphs(intro, please)
        message += toTightUL(conflictPaths)
        raise AbortTask(message)


class StageFiles(_BaseStagingTask):
    def flow(self, deltas: list[GitDelta]):
        if not deltas:  # Nothing to stage (may happen if user keeps pressing Enter in file list view)
            QApplication.beep()
            raise AbortTask()

        self.denyConflicts(deltas, PatchPurpose.Stage)

        paths = [d.new.path for d in deltas]

        self.epilog.effects |= TaskEffects.Workdir
        yield from self.flowCallGit("add", "--", *paths)

        yield from self.debriefPostStage(deltas)

        self.epilog.status = _n("File staged.", "{n} files staged.", len(deltas))

    def debriefPostStage(self, deltas: list[GitDelta]):
        debrief = {}

        for delta in deltas:
            # Staging a tree that isn't registered as a submodule
            if delta.new.mode == FileMode.TREE:
                m = _("You’ve added another Git repo inside your current repo. "
                      "It is STRONGLY RECOMMENDED to absorb it as a submodule before committing.")

            # Staging a submodule with uncommitted changes within
            elif delta.new.mode == FileMode.COMMIT and delta.submoduleWorkdirDirty:
                m = _("Uncommitted changes in the submodule can’t be staged from the parent repository.")

            # Staging a submodule deletion, and the submodule is still in .gitmodules
            elif (delta.status == GitStatus.Deleted
                  and delta.old.mode == FileMode.COMMIT
                  and delta.old.path in self.repo.listall_submodules_dict_at_head()):
                m = _("Don’t forget to remove the submodule from {0} to complete its deletion.", tquo(DOT_GITMODULES))

            # Otherwise, no special message
            else:
                continue

            debrief[delta.new.path] = m

        if not debrief:
            return

        # For better perceived responsivity, show message box asynchronously
        # so that RefreshRepo occurs in the background after the task completes
        yield from self.flowEnterUiThread()
        qmb = asyncMessageBox(
            self.parentWidget(),
            'information',
            self.name(),
            _n("An item requires your attention after staging:", "{n} items require your attention after staging:", len(debrief)))
        addULToMessageBox(qmb, [f"{btag(path)}: {issue}" for path, issue in debrief.items()])
        qmb.show()


class DiscardFiles(_BaseStagingTask):
    def flow(self, deltas: list[GitDelta]):
        assert all(d.source == GitDeltaSource.Dirty for d in deltas)

        textPara = []
        really = ""
        verb = _("Discard changes")

        if not deltas:  # Nothing to discard (may happen if user keeps pressing Delete in file list view)
            QApplication.beep()
            raise AbortTask()

        self.denyConflicts(deltas, PatchPurpose.Discard)

        # Filter on submodules
        submoPaths = [d.new.path for d in deltas if d.isSubtreeCommitPatch()]
        numSubmos = len(submoPaths)

        if len(deltas) == 1:
            delta = deltas[0]
            bpath = bquo(delta.new.path)
            if delta.status == GitStatus.Untracked:
                really = _("Really delete {0}?", bpath)
                really += " " + _("Git isn’t tracking this file, so you may not be able to recover it from older commits.")
                verb = _("Delete")
            elif delta.new.mode == FileMode.COMMIT:  # modeWorktree
                really = _("Really discard changes in submodule {0}?", bpath)
            else:
                really = _("Really discard changes to {0}?", bpath)
        else:
            numFiles = len(deltas) - numSubmos
            if numSubmos and numFiles:
                really = _("Really discard changes to {nf} files and in {ns} submodules?", nf=numFiles, ns=numSubmos)
            elif numSubmos:
                really = _("Really discard changes in {n} submodules?", n=numSubmos)
            else:
                really = _("Really discard changes to {n} files?", n=numFiles)

        textPara.append(really)
        if numSubmos:
            submoPostamble = _n(
                "Any uncommitted changes in the submodule will be <b>cleared</b> and the submodule’s HEAD will be reset.",
                "Any uncommitted changes in {n} submodules will be <b>cleared</b> and the submodules’ HEAD will be reset.",
                numSubmos)
            textPara.append(submoPostamble)

        textPara.append(_("This cannot be undone!"))

        yield from self.flowConfirm(text=paragraphs(*textPara), verb=verb, buttonIcon="git-discard")

        self.epilog.effects |= TaskEffects.Workdir
        if numSubmos:
            self.epilog.effects |= TaskEffects.Refs  # We don't have TaskEffects.Submodules so .Refs is the next best thing

        # Back up discarded patches
        if Trash.enabled():
            for delta in deltas:
                try:
                    yield from self._backupDelta(delta)
                except Trash.BackupSkipped as ex:
                    logger.warning(f"Backup skipped: {ex}")

        tracked = [d.new.path for d in deltas if d.status != GitStatus.Untracked]
        untrackedFiles = [d.new.path for d in deltas if d.status == GitStatus.Untracked and d.new.mode != FileMode.TREE]
        untrackedTrees = [d.new.path for d in deltas if d.status == GitStatus.Untracked and d.new.mode == FileMode.TREE]

        # Discard untracked trees. They have already been backed up above,
        # but restore_files_from_index isn't capable of removing trees.
        for untrackedTree in untrackedTrees:
            untrackedTreePath = self.repo.in_workdir(untrackedTree)
            assert os.path.isdir(untrackedTreePath)
            shutil.rmtree(untrackedTreePath)

        if untrackedFiles:
            yield from self.flowCallGit("clean", "--force", "--", *untrackedFiles)
        if tracked:
            yield from self.flowCallGit("checkout", "--", *tracked)

        if submoPaths:
            for submo in submoPaths:
                subWd = os.path.join(self.repo.workdir, submo)
                yield from self.flowCallGit("clean", "-d", "--force", workdir=subWd)
            yield from self.flowCallGit("submodule", "update", "--force", "--init", "--recursive", "--checkout", "--", *submoPaths)

        self.epilog.status = _n("File discarded.", "{n} files discarded.", len(deltas))

    def _backupDelta(self, delta: GitDelta):
        # Don't back up deletions
        if delta.status == GitStatus.Deleted:
            return

        trash = Trash.instance()
        path = delta.new.path
        workdir = self.repo.workdir

        if delta.status == GitStatus.Untracked:
            if delta.new.mode == FileMode.TREE:
                # Untracked tree
                trash.backupTree(workdir, path)
            else:
                # Untracked file
                # TODO: Also binary files?
                trash.backupFile(workdir, path)
        else:
            # TODO: Cache patch in GitDelta? So we don't have to regenerate the patch if we've already displayed it
            tokens = GitDriver.buildDiffCommand(delta)
            driver = yield from self.flowCallGit(*tokens, autoFail=False)
            patchText = driver.stdoutScrollback()
            trash.backupPatch(workdir, patchText, path)


class UnstageFiles(_BaseStagingTask):
    def flow(self, deltas: list[GitDelta]):
        if not deltas:  # Nothing to unstage (may happen if user keeps pressing Delete in file list view)
            QApplication.beep()
            raise AbortTask()

        paths = []
        for delta in deltas:
            paths.append(delta.new.path)
            if delta.status == GitStatus.Renamed:
                paths.append(delta.old.path)

        self.epilog.effects |= TaskEffects.Workdir
        # Not using 'restore --staged' because it doesn't work in an empty repo
        yield from self.flowCallGit("reset", "--", *paths)

        self.epilog.status = _n("File unstaged.", "{n} files unstaged.", len(deltas))


class DiscardModeChanges(_BaseStagingTask):
    def flow(self, deltas: list[GitDelta]):
        paths = [delta.new.path for delta in deltas]
        numFiles = len(paths)

        if numFiles == 0:  # Nothing to unstage (may happen if user keeps pressing Delete in file list view)
            QApplication.beep()
            raise AbortTask()

        text = paragraphs(
            (_("Really discard mode change in {0}?", bquo(paths[0]))
             if numFiles == 1 else
             _("Really discard mode changes in <b>{n} files</b>?", n=numFiles)),
            _("This cannot be undone!"))

        yield from self.flowConfirm(text=text, verb=_("Discard mode changes"), buttonIcon="git-discard")

        yield from self.flowEnterWorkerThread()
        self.epilog.effects |= TaskEffects.Workdir
        self.repo.discard_mode_changes(paths)


class UnstageModeChanges(_BaseStagingTask):
    def flow(self, deltas: list[GitDelta]):
        if not deltas:  # Nothing to unstage (may happen if user keeps pressing Delete in file list view)
            QApplication.beep()
            raise AbortTask()

        yield from self.flowEnterWorkerThread()
        self.epilog.effects |= TaskEffects.Workdir

        self.repo.refresh_index()
        index = self.repo.index
        for delta in deltas:
            of = delta.old
            nf = delta.new
            if (of.mode != nf.mode
                    and delta.status not in [GitStatus.Added, GitStatus.Deleted, GitStatus.Untracked]
                    and of.mode in [FileMode.BLOB, FileMode.BLOB_EXECUTABLE]):
                index.add(IndexEntry(nf.path, Oid(hex=nf.id), of.mode))
        index.write()


class ApplyPatch(RepoTask):
    def flow(self, delta: GitDelta, subPatch: str, purpose: PatchPurpose):
        if not subPatch:
            QApplication.beep()
            verb = trtables.enum(purpose & PatchPurpose.VerbMask).lower()
            message = _("Can’t {verb} the selection because no red/green lines are selected.", verb=verb)
            raise AbortTask(message, asStatusMessage=True)

        if purpose & PatchPurpose.Discard:
            title = trtables.enum(purpose)
            isHunk = purpose & PatchPurpose.Hunk
            text = paragraphs(
                _("Really discard this hunk?") if isHunk else _("Really discard the selected lines?"),
                _("This cannot be undone!"))
            yield from self.flowConfirm(title, text=text, verb=title, buttonIcon="git-discard-lines")

            try:
                Trash.instance().backupPatch(self.repo.workdir, subPatch, delta.new.path)
            except Trash.BackupSkipped as ex:
                logger.warning(f"Backup skipped: {ex}")

            applyLocation = ApplyLocation.WORKDIR
        else:
            applyLocation = ApplyLocation.INDEX

        yield from self.flowEnterWorkerThread()
        self.epilog.effects |= TaskEffects.Workdir

        # libgit2's index must be fresh in order to apply a partial patch to the index.
        if applyLocation == ApplyLocation.INDEX:
            self.repo.refresh_index()

        self.repo.apply(subPatch, applyLocation)

        self.epilog.status = ApplyPatch.patchPurposePastTense(purpose)

    @staticmethod
    def patchPurposePastTense(purpose: PatchPurpose):
        pp = PatchPurpose
        t = {
            pp.Lines | pp.Stage: _p("PatchPurpose", "Lines staged."),
            pp.Lines | pp.Unstage: _p("PatchPurpose", "Lines unstaged."),
            pp.Lines | pp.Discard: _p("PatchPurpose", "Lines discarded."),
            pp.Hunk | pp.Stage: _p("PatchPurpose", "Hunk staged."),
            pp.Hunk | pp.Unstage: _p("PatchPurpose", "Hunk unstaged."),
            pp.Hunk | pp.Discard: _p("PatchPurpose", "Hunk discarded."),
            pp.File | pp.Stage: _n("File staged.", "{n} files staged.", 1),
            pp.File | pp.Unstage: _n("File unstaged.", "{n} files unstaged.", 1),
            pp.File | pp.Discard: _n("File discarded.", "{n} files discarded.", 1),
        }
        return t.get(purpose, "")


class HardSolveConflicts(RepoTask):
    def flow(self, conflicts: list[GitConflict], keepOurs: bool):
        # First ask for confirmation
        yield from self._confirm(conflicts, keepOurs)

        # Sort files in 'keep' and 'nuke' buckets
        files = [c.ours if keepOurs else c.theirs for c in conflicts]
        keepPaths = [f.path for f in files if not f.isId0()]
        nukePaths = [f.path for f in files if f.isId0()]

        # Back up nuked files + THEIRS
        if Trash.enabled():
            backupList = nukePaths[:]
            if not keepOurs:
                backupList += keepPaths

            for path in backupList:
                try:
                    Trash.instance().backupFile(self.repo.workdir, path)
                except Trash.BackupSkipped as ex:
                    logger.warning(f"Backup skipped: {ex}")

        self.epilog.effects |= TaskEffects.Workdir

        # Restore desired sides, then stage the files to resolve the conflict
        if keepPaths:
            sideArg = "--ours" if keepOurs else "--theirs"
            yield from self.flowCallGit("restore", "--progress", sideArg, "--", *keepPaths)
            yield from self.flowCallGit("add", "--force", "--", *keepPaths)

        # Stage deletions to resolve the conflict
        if nukePaths:
            yield from self.flowCallGit("rm", "--", *nukePaths)

        # Jump to any staged file after the task
        for path in keepPaths:
            if Path(self.repo.in_workdir(path)).is_file():
                self.epilog.jumpTo = NavLocator.inStaged(path)
                break

        self.epilog.status = _n("Conflict resolved.", "{n} conflicts resolved.", len(conflicts))

    def _confirm(self, conflicts: list[GitConflict], keepOurs: bool):
        title = _n("Resolve conflict", "Resolve {n} conflicts", len(conflicts))
        verb = _("Keep OUR version") if keepOurs else _("Accept THEIR version")
        promptSuffix = ""

        sidesSet = {c.sides for c in conflicts}
        if len(sidesSet) == 1:
            from gitfourchette.forms.conflictview import GitConflictSidesLocalization
            strings = GitConflictSidesLocalization.getStrings(conflicts[0].sides)
            a = _("Reject incoming changes") if keepOurs else _("Accept incoming changes")
            b = strings.ours2 if keepOurs else strings.theirs2
            promptSuffix = f"<p>\u2192 {a}.<br>\u2192 {b}."

        if len(conflicts) == 1:
            if keepOurs:
                singlePath = hquoe(Path(conflicts[0].ours.path).name)
            else:
                singlePath = hquoe(Path(conflicts[0].theirs.path).name)
            prompt = _("<b>{verb}</b> to resolve the conflict on {single}?",
                       verb=verb, single=singlePath)
        else:
            prompt = _n("<b>{verb}</b> to resolve the conflict on {n} file?",
                        "<b>{verb}</b> to resolve conflicts on {n} files?",
                        n=len(conflicts), verb=verb)

        yield from self.flowConfirm(title, text=prompt + promptSuffix, verb=verb)


class OpenMergeTool(RepoTask):
    def flow(self, conflict: GitConflict, reopenWorkInProgress: bool):
        mergeDriver = MergeDriver.findOngoingMerge(conflict)

        if mergeDriver is None:
            mergeDriver = yield from self.makeMergeDriver(conflict)

        mergeDriver.startProcess(reopenWorkInProgress)
        return mergeDriver

    def makeMergeDriver(self, conflict: GitConflict):
        mergeTempDir = QTemporaryDir(str(Path(qTempDir(), "merge")))
        into = mergeTempDir.path()
        repo = self.repo

        def dump(df: GitDeltaFile, prefix: str):
            path = yield from self.flowSubtask(SaveDeltaFileAs, df, saveIntoDir=into, prefix=prefix)
            return path

        targetPath = repo.in_workdir(conflict.ours.path)
        baseName = Path(targetPath).name

        # Dump OURS and THEIRS blobs into the temporary directory
        oursPath = yield from dump(conflict.ours, prefix="[OURS]")
        theirsPath = yield from dump(conflict.theirs, prefix="[THEIRS]")

        if conflict.ancestor:
            # Dump ANCESTOR blob into the temporary directory
            ancestorPath = yield from dump(conflict.ancestor, prefix="[ANCESTOR]")
        else:
            # There's no ancestor! Some merge tools can fake a 3-way merge without
            # an ancestor (e.g. PyCharm), but others won't (e.g. VS Code).
            # To make sure we get a 3-way merge, copy our current workdir file as
            # the fake ANCESTOR file. It should contain chevron conflict markers
            # (<<<<<<< >>>>>>>) which should trigger conflicts between OURS and
            # THEIRS in the merge tool.
            ancestorPath = str(Path(into, f"[NO-ANCESTOR]{baseName}"))
            shutil.copyfile(targetPath, ancestorPath)

        # Create scratch file (merge tool output).
        # Some merge tools (such as VS Code) use the contents of this file
        # as a starting point, so copy the workdir version for this purpose.
        scratchPath = Path(into, f"[MERGED]{baseName}")
        shutil.copyfile(targetPath, scratchPath)

        paths = MergeDriver.MergeFiles(
            ours=oursPath,
            theirs=theirsPath,
            ancestor=ancestorPath,
            scratch=str(scratchPath),
            target=targetPath)

        mergeDriver = MergeDriver(self.parentWidget(), conflict, paths)

        # Keep a reference to temp dir so it doesn't vanish
        mergeDriver._keepAroundMergeTempDir = mergeTempDir  # type: ignore[attr-defined]

        return mergeDriver


class PreviewDeltaFile(RepoTask):
    def flow(self, df: GitDeltaFile, prefix: str, registerCallback: Callable[[CodeWindow], None]):
        # Don't load large files
        maxFileSize = settings.prefs.largeFileThresholdKB * 1024
        if maxFileSize != 0 and GitDeltaFile.SupportsFastSizeBallpark:
            ballpark = df.sizeBallpark(self.repo)
            if maxFileSize < ballpark:
                locale = QLocale()
                humanSize = locale.formattedDataSize(ballpark, 1)
                message = _("This file is very large.") + f" ({humanSize})"
                yield from self.flowConfirm(_p("ConflictView", "Preview"), message, verb=_("Show anyway"))

        # Dump the file
        path = yield from self.flowSubtask(SaveDeltaFileAs, df, saveIntoDir=qTempDir())
        pathObj = Path(path)
        text = pathObj.read_text("utf-8")
        pathObj.unlink()

        ident = hash(text) ^ hash(df.path)

        # Raise existing window, if any
        if CodeWindow.activateExistingWindow(ident):
            return

        codeWindow = CodeWindow(uniqueIdentifier=ident)
        codeWindow.setPlainText(text, df.path)
        codeWindow.setWindowTitle(f"[{prefix}] {Path(df.path).name}")
        codeWindow.show()
        codeWindow.codeView.setFocus()

        self.rw.aboutToDelete.connect(codeWindow.close)

        registerCallback(codeWindow)


class AcceptMergeConflictResolution(RepoTask):
    def canKill(self, task: RepoTask) -> bool:
        from gitfourchette.tasks import RefreshRepo, Jump
        return isinstance(task, RefreshRepo | Jump)

    def flow(self, mergeDriver: MergeDriver):
        self.epilog.effects |= TaskEffects.Workdir
        path = mergeDriver.conflict.ours.path
        mergeDriver.copyScratchToTarget()
        mergeDriver.deleteNow()
        yield from self.flowCallGit("add", "--force", "--", path)

        # Jump to staged file after confirming conflict resolution
        self.epilog.jumpTo = NavLocator.inStaged(path)
        self.epilog.status = _("Merge conflict resolved in {0}.", tquo(path))


class ApplyPatchFile(RepoTask):
    def flow(self, path: str = ""):
        yield from ApplyPatchFile.do(self, path=path)

    @staticmethod
    def do(
            task: RepoTask,
            path: str = "",
            reverse: bool = False,
            context: int = -1,
            title: str = "",
            question: str = "",
    ):
        # Fallback title
        title = title or task.name()

        # If no path, bring up file dialog
        if not path:
            patchFileCaption = _("Patch file")
            allFilesCaption = _("All files")

            qfd = PersistentFileDialog.openFile(
                task.parentWidget(), "OpenPatch", title,
                filter=f"{patchFileCaption} (*.patch);;{allFilesCaption} (*)")
            path = yield from task.flowFileDialog(qfd)

        # Fallback question
        if not question:
            verb = _("revert") if reverse else _("apply")
            basename = Path(path).name
            question = _("Do you want to {verb} patch file {path}?",
                         verb=btag(verb), path=bquoe(basename))

        # Build command
        stem = [
            "apply",
            *argsIf(reverse, "--reverse"),
            *argsIf(context >= 0, f"-C{context}"),
            path
        ]

        # Do a dry run first.
        driver = yield from task.flowCallGit(*stem, "--numstat", "-z", "--check")

        table = driver.stdoutTableNumstatZ()
        numFiles = len(table)
        details = []
        firstFile = ""
        for adds, dels, patchFile in table:
            if adds == "-" or dels == "-":
                details.append(_("(binary)") + " " + escape(patchFile))
            else:
                details.append(f"(<add>+{adds}</add> <del>-{dels}</del>) {escape(patchFile)}")
            firstFile = firstFile or patchFile

        addDelStyle = settings.prefs.addDelColorsStyleTag()
        listIntro = _n("<b>{n}</b> file will be modified in your working directory:",
                       "<b>{n}</b> files will be modified in your working directory:", n=numFiles)
        confirmText = addDelStyle + paragraphs(question, listIntro)

        yield from task.flowConfirm(title, confirmText, verb=_("Apply"), detailList=details)

        # Dry run confirmed, go ahead
        task.epilog.effects |= TaskEffects.Workdir

        yield from task.flowCallGit(*stem)

        task.epilog.jumpTo = NavLocator.inUnstaged(firstFile)
        task.epilog.status = _n("{n} file modified in the working directory.",
                                "{n} files modified in the working directory.", n=numFiles)


class ApplyPatchFileReverse(RepoTask):
    def flow(self, path: str = ""):
        yield from ApplyPatchFile.do(self, path=path, reverse=True)


class ApplyPatchData(RepoTask):
    def flow(self, patchData: str, title: str, question: str, reverse: bool = False, context: int = -1):
        assert isinstance(patchData, str), "patchData should be str"

        if not patchData:
            raise AbortTask(_("There’s nothing to apply in the selection."))

        template = os.path.join(qTempDir(), self.__class__.__name__ + "-XXXXXX.patch")
        tempPatch = QTemporaryFile(template, self)
        tempPatch.open()
        tempPatch.write(patchData.encode("utf-8"))
        tempPatch.close()
        path = tempPatch.fileName()

        yield from ApplyPatchFile.do(
            self,
            path=path,
            reverse=reverse,
            context=context,
            title=title,
            question=question)


class RestoreRevisionToWorkdir(RepoTask):
    def flow(self, delta: GitDelta, old: bool):
        if old:
            preposition = _p("preposition slotted into '...BEFORE this commit'", "before")
            diffFile = delta.old
            delete = delta.status == GitStatus.Added
        else:
            preposition = _p("preposition slotted into '...AT this commit'", "at")
            diffFile = delta.new
            delete = delta.status == GitStatus.Deleted

        path = self.repo.in_workdir(diffFile.path)
        pathObj = Path(path)
        existsNow = pathObj.is_file()

        if not existsNow and delete:
            message = _("Your working copy of {path} already matches the revision {preposition} this commit.",
                        path=bquo(diffFile.path), preposition=preposition)
            raise AbortTask(message, icon="information")

        if not existsNow:
            actionVerb = _("recreated")
        elif delete:
            actionVerb = _("deleted")
        else:
            actionVerb = _("overwritten")
        prompt = paragraphs(
            _("Do you want to restore {path} as it was {preposition} this commit?",
              path=bquo(diffFile.path), preposition=preposition),
            _("This file will be {processed} in your working directory.", processed=actionVerb))

        yield from self.flowConfirm(text=prompt, verb=_("Restore"))

        self.epilog.effects |= TaskEffects.Workdir

        if delete:
            pathObj.unlink()
        else:
            assert diffFile.sourceCommit not in [None, NULL_OID]
            yield from self.flowCallGit(
                "restore",
                "--progress",
                f"--source={diffFile.sourceCommit}",
                "--",
                diffFile.path)

        self.epilog.status = _("File {path} {processed}.", path=tquoe(diffFile.path), processed=actionVerb)
        self.epilog.jumpTo = NavLocator.inUnstaged(diffFile.path)


class SaveDeltaFileAs(RepoTask):
    def flow(self, file: GitDeltaFile, saveIntoDir: str = "", prefix: str = ""):
        dfPath = Path(file.path)
        suggStem = prefix + dfPath.stem

        if file.source == GitDeltaSource.Commit:
            assert file.sourceCommit not in [None, NULL_OID]
            suggStem += f"@{shortHash(file.sourceCommit)}"
            catFileArgs = ["--filters", f"{file.sourceCommit}:{file.path}"]  # --filters for LFS awareness
        elif file.source == GitDeltaSource.Index:
            catFileArgs = ["--filters", f":{file.path}"]  # --filters for LFS awareness
        elif file.source == GitDeltaSource.Unknown:  # Most likely from GitConflict
            assert file.isIdValid()
            catFileArgs = ["blob", str(file.id)]
        else:
            raise NotImplementedError()

        suggExt = dfPath.suffix
        suggName = suggStem + suggExt

        if saveIntoDir:
            targetStr = withUniqueSuffix(suggStem, ext=suggExt, reserved=lambda s: Path(saveIntoDir, s).exists())
            targetStr = str(Path(saveIntoDir, targetStr))
        else:
            qfd = PersistentFileDialog.saveFile(self.parentWidget(), "SaveFile", _("Save file revision as"), suggName)
            targetStr = yield from self.flowFileDialog(qfd)
        target = Path(targetStr)

        driver = yield from self.flowCallGit("cat-file", *catFileArgs)

        assert not driver._stdout, "stdout consumed prematurely"
        data = driver.readAllStandardOutput().data()
        target.write_bytes(data)

        if file.mode == FileMode.BLOB_EXECUTABLE:
            mode = 0o100 | target.lstat().st_mode
            target.lchmod(mode)

        return str(target)


class SaveRevisionAs(RepoTask):
    def flow(self, delta: GitDelta, old: bool, saveIntoDir: str = "", prefix: str = ""):
        if old:
            diffFile = delta.old
            if delta.status == GitStatus.Added:
                raise AbortTask(_("This file didn’t exist before the commit."))
        else:
            diffFile = delta.new
            if delta.status == GitStatus.Deleted:
                raise AbortTask(_("This file was deleted by the commit."))

        outPath = yield from self.flowSubtask(SaveDeltaFileAs, diffFile, saveIntoDir, prefix)
        return outPath


class OpenRevisionInEditor(RepoTask):
    def flow(self, delta: GitDelta, old: bool):
        path = yield from self.flowSubtask(SaveRevisionAs, delta, old, qTempDir())
        assert isinstance(path, str)

        ToolProcess.startTextEditor(self.parentWidget(), path)


class OpenInDiffTool(RepoTask):
    def flow(self, delta: GitDelta):
        if delta.new.isId0():
            raise FileNotFoundError(_("Can’t open external diff tool on a deleted file."))

        if delta.old.isId0():
            raise FileNotFoundError(_("Can’t open external diff tool on a new file."))

        into = qTempDir()

        def dump(file: GitDeltaFile, prefix: str):
            path = yield from self.flowSubtask(SaveDeltaFileAs, file, saveIntoDir=into, prefix=prefix)
            return path

        if delta.source == GitDeltaSource.Dirty:
            # Unstaged: compare indexed state to workdir file
            oldPath = yield from dump(delta.old, "[INDEXED]")
            newPath = self.repo.in_workdir(delta.new.path)
        elif delta.source == GitDeltaSource.Index:
            # Staged: compare HEAD state to indexed state
            oldPath = yield from dump(delta.old, "[HEAD]")
            newPath = yield from dump(delta.new, "[STAGED]")
        elif delta.source == GitDeltaSource.Commit:
            # Committed: compare parent state to this commit
            oldPath = yield from dump(delta.old, "[OLD]")
            newPath = yield from dump(delta.new, "[NEW]")
        else:
            raise NotImplementedError(f"unsupported source {delta.source}")

        return ToolProcess.startDiffTool(self.parentWidget(), oldPath, newPath)


class AbortMerge(RepoTask):
    def flow(self):
        self.repo.refresh_index()

        isMerging = self.repo.state() == RepositoryState.MERGE
        isCherryPicking = self.repo.state() == RepositoryState.CHERRYPICK
        isReverting = self.repo.state() == RepositoryState.REVERT
        anyConflicts = self.repo.index.conflicts

        if not (isMerging or isCherryPicking or isReverting or anyConflicts):
            raise AbortTask(_("No abortable state is in progress."), icon='information')

        if isCherryPicking:
            clause = _("abort the ongoing cherry-pick")
            title = _("Abort cherry-pick")
            postStatus = _("Cherry-pick aborted.")
            gitCommand = ["cherry-pick", "--abort"]
        elif isMerging:
            clause = _("abort the ongoing merge")
            title = _("Abort merge")
            postStatus = _("Merge aborted.")
            gitCommand = ["merge", "--abort"]
        elif isReverting:
            clause = _("abort the ongoing revert")
            title = _("Abort revert")
            postStatus = _("Revert aborted.")
            gitCommand = ["revert", "--abort"]
        else:
            clause = _("reset the index")
            title = _("Reset index")
            postStatus = _("Index reset.")
            gitCommand = ["reset", "--merge"]

        try:
            abortList = self.repo.get_reset_merge_file_list()
        except MultiFileError as exc:
            exc.message = _n(
                "Cannot {verb} right now, because a file contains both staged and unstaged changes.",
                "Cannot {verb} right now, because {n} files contain both staged and unstaged changes.",
                n=len(exc.file_exceptions), verb=clause)
            exc.message += " " + _("Please unstage the changes and try again.")
            raise  # re-raise MultiFileError

        lines = [_("Do you want to {0}?", clause)]

        if not abortList:
            lines.append(_("No files are affected."))
        else:
            if anyConflicts:
                lines.append(_("All conflicts will be cleared and all <b>staged</b> changes will be lost."))
            else:
                lines.append(_("All <b>staged</b> changes will be lost."))
            lines.append(_n("This file will be reset:", "{n} files will be reset:", len(abortList)))

        yield from self.flowConfirm(title=title, text=paragraphs(*lines), verb=englishTitleCase(title),
                                    detailList=[escape(f) for f in abortList])

        self.epilog.effects |= TaskEffects.DefaultRefresh

        yield from self.flowCallGit(*gitCommand)
        # self.repo.reset_merge()
        # self.repo.state_cleanup()

        self.epilog.status = postStatus

        # Clear draft commit message that was set in CherrypickCommit/RevertCommit/MergeBranch
        if isCherryPicking or isReverting or isMerging:
            self.repoModel.prefs.clearDraftCommit()



# Form implementation generated from reading ui file 'conflictview.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_ConflictView(object):
    def setupUi(self, ConflictView):
        ConflictView.setObjectName("ConflictView")
        ConflictView.resize(600, 400)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(ConflictView.sizePolicy().hasHeightForWidth())
        ConflictView.setSizePolicy(sizePolicy)
        self.mainLayout = QGridLayout(ConflictView)
        self.mainLayout.setObjectName("mainLayout")
        self.explainer = QLabel(parent=ConflictView)
        self.explainer.setText("<b>BLURB</b><br>BLURB")
        self.explainer.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.explainer.setWordWrap(True)
        self.explainer.setObjectName("explainer")
        self.mainLayout.addWidget(self.explainer, 3, 0, 1, 1)
        spacerItem = QSpacerItem(20, 20, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        self.mainLayout.addItem(spacerItem, 1, 0, 1, 1)
        self.titleLabel = QLabel(parent=ConflictView)
        self.titleLabel.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.titleLabel.setWordWrap(True)
        self.titleLabel.setObjectName("titleLabel")
        self.mainLayout.addWidget(self.titleLabel, 0, 0, 1, 1)
        self.frame = QFrame(parent=ConflictView)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.frame.sizePolicy().hasHeightForWidth())
        self.frame.setSizePolicy(sizePolicy)
        self.frame.setMaximumSize(QSize(512, 16777215))
        self.frame.setObjectName("frame")
        self.frameLayout = QGridLayout(self.frame)
        self.frameLayout.setContentsMargins(0, 0, 0, 0)
        self.frameLayout.setObjectName("frameLayout")
        self.oursTitleLayout = QHBoxLayout()
        self.oursTitleLayout.setObjectName("oursTitleLayout")
        spacerItem1 = QSpacerItem(1, 1, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.oursTitleLayout.addItem(spacerItem1)
        self.oursIcon = QLabel(parent=self.frame)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.oursIcon.sizePolicy().hasHeightForWidth())
        self.oursIcon.setSizePolicy(sizePolicy)
        self.oursIcon.setText("ICON")
        self.oursIcon.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.oursIcon.setObjectName("oursIcon")
        self.oursTitleLayout.addWidget(self.oursIcon)
        self.oursLabel = QLabel(parent=self.frame)
        self.oursLabel.setEnabled(False)
        self.oursLabel.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.oursLabel.setObjectName("oursLabel")
        self.oursTitleLayout.addWidget(self.oursLabel)
        self.oursPreviewButton = QToolButton(parent=self.frame)
        self.oursPreviewButton.setAutoRaise(True)
        self.oursPreviewButton.setObjectName("oursPreviewButton")
        self.oursTitleLayout.addWidget(self.oursPreviewButton)
        spacerItem2 = QSpacerItem(1, 1, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.oursTitleLayout.addItem(spacerItem2)
        self.frameLayout.addLayout(self.oursTitleLayout, 0, 0, 1, 1)
        self.theirsButton = QPushButton(parent=self.frame)
        self.theirsButton.setMinimumSize(QSize(0, 48))
        self.theirsButton.setText("THEIRS")
        self.theirsButton.setObjectName("theirsButton")
        self.frameLayout.addWidget(self.theirsButton, 1, 2, 1, 1)
        self.orLabel = QLabel(parent=self.frame)
        self.orLabel.setEnabled(False)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.orLabel.sizePolicy().hasHeightForWidth())
        self.orLabel.setSizePolicy(sizePolicy)
        self.orLabel.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.orLabel.setObjectName("orLabel")
        self.frameLayout.addWidget(self.orLabel, 1, 1, 1, 1)
        self.oursArrow = QLabel(parent=self.frame)
        self.oursArrow.setEnabled(False)
        self.oursArrow.setText("↘")
        self.oursArrow.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.oursArrow.setObjectName("oursArrow")
        self.frameLayout.addWidget(self.oursArrow, 2, 0, 1, 1)
        self.theirsTitleLayout = QHBoxLayout()
        self.theirsTitleLayout.setObjectName("theirsTitleLayout")
        spacerItem3 = QSpacerItem(1, 1, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.theirsTitleLayout.addItem(spacerItem3)
        self.theirsIcon = QLabel(parent=self.frame)
        self.theirsIcon.setText("ICON")
        self.theirsIcon.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.theirsIcon.setObjectName("theirsIcon")
        self.theirsTitleLayout.addWidget(self.theirsIcon)
        self.theirsLabel = QLabel(parent=self.frame)
        self.theirsLabel.setEnabled(False)
        self.theirsLabel.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.theirsLabel.setObjectName("theirsLabel")
        self.theirsTitleLayout.addWidget(self.theirsLabel)
        self.theirsPreviewButton = QToolButton(parent=self.frame)
        self.theirsPreviewButton.setAutoRaise(True)
        self.theirsPreviewButton.setObjectName("theirsPreviewButton")
        self.theirsTitleLayout.addWidget(self.theirsPreviewButton)
        spacerItem4 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.theirsTitleLayout.addItem(spacerItem4)
        self.frameLayout.addLayout(self.theirsTitleLayout, 0, 2, 1, 1)
        self.theirsArrow = QLabel(parent=self.frame)
        self.theirsArrow.setEnabled(False)
        self.theirsArrow.setText("↙")
        self.theirsArrow.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.theirsArrow.setObjectName("theirsArrow")
        self.frameLayout.addWidget(self.theirsArrow, 2, 2, 1, 1)
        self.oursButton = QPushButton(parent=self.frame)
        self.oursButton.setMinimumSize(QSize(0, 48))
        self.oursButton.setText("OURS")
        self.oursButton.setObjectName("oursButton")
        self.frameLayout.addWidget(self.oursButton, 1, 0, 1, 1)
        self.stackedWidget = QStackedWidget(parent=self.frame)
        self.stackedWidget.setObjectName("stackedWidget")
        self.mergePage = QWidget()
        self.mergePage.setObjectName("mergePage")
        self.mergeLayout = QGridLayout(self.mergePage)
        self.mergeLayout.setContentsMargins(0, 0, 0, 0)
        self.mergeLayout.setObjectName("mergeLayout")
        self.mergeToolButton = QToolButton(parent=self.mergePage)
        self.mergeToolButton.setAutoRaise(True)
        self.mergeToolButton.setObjectName("mergeToolButton")
        self.mergeLayout.addWidget(self.mergeToolButton, 2, 1, 1, 1)
        self.mergeButton = QPushButton(parent=self.mergePage)
        self.mergeButton.setMinimumSize(QSize(0, 48))
        self.mergeButton.setObjectName("mergeButton")
        self.mergeLayout.addWidget(self.mergeButton, 0, 0, 1, 2)
        self.mergeToolStatus = QLabel(parent=self.mergePage)
        self.mergeToolStatus.setText("MERGE TOOL STATUS")
        self.mergeToolStatus.setWordWrap(True)
        self.mergeToolStatus.setObjectName("mergeToolStatus")
        self.mergeLayout.addWidget(self.mergeToolStatus, 2, 0, 1, 1)
        spacerItem5 = QSpacerItem(20, 8, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.mergeLayout.addItem(spacerItem5, 4, 1, 1, 1)
        self.stackedWidget.addWidget(self.mergePage)
        self.mergeInProgressPage = QWidget()
        self.mergeInProgressPage.setObjectName("mergeInProgressPage")
        self.mergeInProgressLayout = QGridLayout(self.mergeInProgressPage)
        self.mergeInProgressLayout.setContentsMargins(0, 0, 0, 0)
        self.mergeInProgressLayout.setObjectName("mergeInProgressLayout")
        spacerItem6 = QSpacerItem(20, 8, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.mergeInProgressLayout.addItem(spacerItem6, 3, 0, 1, 1)
        self.progressBar = QProgressBar(parent=self.mergeInProgressPage)
        self.progressBar.setMaximum(0)
        self.progressBar.setProperty("value", 0)
        self.progressBar.setObjectName("progressBar")
        self.mergeInProgressLayout.addWidget(self.progressBar, 1, 0, 1, 1)
        self.cancelMergeInProgress = QPushButton(parent=self.mergeInProgressPage)
        self.cancelMergeInProgress.setObjectName("cancelMergeInProgress")
        self.mergeInProgressLayout.addWidget(self.cancelMergeInProgress, 1, 1, 1, 1)
        self.mergeInProgressLabel = QLabel(parent=self.mergeInProgressPage)
        self.mergeInProgressLabel.setText("MERGE IN PROGRESS")
        self.mergeInProgressLabel.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.mergeInProgressLabel.setWordWrap(True)
        self.mergeInProgressLabel.setObjectName("mergeInProgressLabel")
        self.mergeInProgressLayout.addWidget(self.mergeInProgressLabel, 0, 0, 1, 2)
        self.stackedWidget.addWidget(self.mergeInProgressPage)
        self.mergeCompletePage = QWidget()
        self.mergeCompletePage.setObjectName("mergeCompletePage")
        self.mergeCompleteLayout = QGridLayout(self.mergeCompletePage)
        self.mergeCompleteLayout.setContentsMargins(0, 0, 0, 0)
        self.mergeCompleteLayout.setObjectName("mergeCompleteLayout")
        spacerItem7 = QSpacerItem(20, 8, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.mergeCompleteLayout.addItem(spacerItem7, 6, 1, 1, 1)
        self.discardMergeButton = QPushButton(parent=self.mergeCompletePage)
        self.discardMergeButton.setObjectName("discardMergeButton")
        self.mergeCompleteLayout.addWidget(self.discardMergeButton, 3, 0, 1, 1)
        self.confirmMergeLabel = QLabel(parent=self.mergeCompletePage)
        self.confirmMergeLabel.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.confirmMergeLabel.setWordWrap(True)
        self.confirmMergeLabel.setObjectName("confirmMergeLabel")
        self.mergeCompleteLayout.addWidget(self.confirmMergeLabel, 0, 0, 1, 2)
        self.reworkMergeButton = QPushButton(parent=self.mergeCompletePage)
        self.reworkMergeButton.setObjectName("reworkMergeButton")
        self.mergeCompleteLayout.addWidget(self.reworkMergeButton, 3, 1, 1, 1)
        self.confirmMergeButton = QPushButton(parent=self.mergeCompletePage)
        self.confirmMergeButton.setMinimumSize(QSize(0, 48))
        self.confirmMergeButton.setDefault(True)
        self.confirmMergeButton.setObjectName("confirmMergeButton")
        self.mergeCompleteLayout.addWidget(self.confirmMergeButton, 1, 0, 1, 2)
        self.stackedWidget.addWidget(self.mergeCompletePage)
        self.confirmDeletionPage = QWidget()
        self.confirmDeletionPage.setObjectName("confirmDeletionPage")
        self.confirmDeletionLayout = QGridLayout(self.confirmDeletionPage)
        self.confirmDeletionLayout.setContentsMargins(0, 0, 0, 0)
        self.confirmDeletionLayout.setObjectName("confirmDeletionLayout")
        self.confirmDeletionButton = QPushButton(parent=self.confirmDeletionPage)
        self.confirmDeletionButton.setObjectName("confirmDeletionButton")
        self.confirmDeletionLayout.addWidget(self.confirmDeletionButton, 0, 0, 1, 1)
        spacerItem8 = QSpacerItem(20, 8, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.confirmDeletionLayout.addItem(spacerItem8, 1, 0, 1, 1)
        self.stackedWidget.addWidget(self.confirmDeletionPage)
        self.emptyPage = QWidget()
        self.emptyPage.setObjectName("emptyPage")
        self.emptyPageLayout = QVBoxLayout(self.emptyPage)
        self.emptyPageLayout.setContentsMargins(0, 0, 0, 0)
        self.emptyPageLayout.setSpacing(0)
        self.emptyPageLayout.setObjectName("emptyPageLayout")
        spacerItem9 = QSpacerItem(20, 8, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.emptyPageLayout.addItem(spacerItem9)
        self.stackedWidget.addWidget(self.emptyPage)
        self.frameLayout.addWidget(self.stackedWidget, 3, 0, 1, 3)
        self.frameLayout.setColumnStretch(0, 1)
        self.frameLayout.setColumnStretch(2, 1)
        self.mainLayout.addWidget(self.frame, 6, 0, 1, 1)
        spacerItem10 = QSpacerItem(20, 16, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Maximum)
        self.mainLayout.addItem(spacerItem10, 5, 0, 1, 1)

        self.retranslateUi(ConflictView)
        self.stackedWidget.setCurrentIndex(0)
        ConflictView.setTabOrder(self.oursButton, self.theirsButton)
        ConflictView.setTabOrder(self.theirsButton, self.mergeButton)
        ConflictView.setTabOrder(self.mergeButton, self.mergeToolButton)
        ConflictView.setTabOrder(self.mergeToolButton, self.cancelMergeInProgress)
        ConflictView.setTabOrder(self.cancelMergeInProgress, self.confirmMergeButton)
        ConflictView.setTabOrder(self.confirmMergeButton, self.discardMergeButton)
        ConflictView.setTabOrder(self.discardMergeButton, self.reworkMergeButton)
        ConflictView.setTabOrder(self.reworkMergeButton, self.confirmDeletionButton)

    def retranslateUi(self, ConflictView):
        ConflictView.setWindowTitle(_p("ConflictView", "Merge conflict"))
        self.titleLabel.setText(_p("ConflictView", "Merge conflict on {0}"))
        self.oursLabel.setText(_p("ConflictView", "OUR version"))
        self.oursPreviewButton.setToolTip(_p("ConflictView", "Preview"))
        self.oursPreviewButton.setText(_p("ConflictView", "Preview"))
        self.orLabel.setText(_p("ConflictView", "or"))
        self.theirsLabel.setText(_p("ConflictView", "THEIR version"))
        self.theirsPreviewButton.setToolTip(_p("ConflictView", "Preview"))
        self.theirsPreviewButton.setText(_p("ConflictView", "Preview"))
        self.mergeToolButton.setToolTip(_p("ConflictView", "Select another external merge tool in the Settings."))
        self.mergeToolButton.setText(_p("ConflictView", "Change merge tool…"))
        self.mergeButton.setToolTip(_p("ConflictView", "<p>Open {tool} so you can merge both versions by hand. When you are done, save the file and quit {tool} to resolve the conflict."))
        self.mergeButton.setText(_p("ConflictView", "Merge both versions in {tool}"))
        self.cancelMergeInProgress.setText(_p("ConflictView", "Abort"))
        self.discardMergeButton.setToolTip(_p("ConflictView", "<p>Ignore the merge that you made in {tool} so that you can start over."))
        self.discardMergeButton.setText(_p("ConflictView", "Discard this merge"))
        self.confirmMergeLabel.setText(_p("ConflictView", "It looks like you’ve finished merging this file."))
        self.reworkMergeButton.setToolTip(_p("ConflictView", "<p>Merge this file in {tool} again, starting where you left off."))
        self.reworkMergeButton.setText(_p("ConflictView", "Merge again"))
        self.confirmMergeButton.setToolTip(_p("ConflictView", "<p>Resolve this conflict by accepting the merge that you’ve completed in {tool}."))
        self.confirmMergeButton.setText(_p("ConflictView", "Resolve the conflict with your merge"))
        self.confirmDeletionButton.setText(_p("ConflictView", "Confirm deletion"))



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import weakref

from gitfourchette import settings
from gitfourchette.application import GFApplication
from gitfourchette.qt import *


class AutoHideMenuBar(QObject):
    def __init__(self, menuBar: QMenuBar):
        super().__init__(menuBar)

        self.setObjectName("AutoHideMenuBar")

        self.menuBar = menuBar
        self.defaultMaximumHeight = self.menuBar.maximumHeight()
        self.sticky = False

        self.hideScheduler = QTimer(self)
        self.hideScheduler.setInterval(0)
        self.hideScheduler.setSingleShot(True)
        self.hideScheduler.timeout.connect(self.doScheduledHide)

        self.menusConnected: list[weakref.ReferenceType[QMenu]] = []

        GFApplication.instance().prefsChanged.connect(self.refreshPrefs)
        self.refreshPrefs()

    def refreshPrefs(self):
        self.hideScheduler.stop()

        # Show the menu bar if auto-hide is turned off, OR if the sticky bit
        # was set as the prefs were reset. That way, if our specific setting
        # wasn't updated, the menu bar will stay put.
        self.showMenuBar(not self.enabled or self.sticky)

        # Reinstall signal connections regardless.
        self.reconnectToMenus()

        # Reset sticky bit if auto-hide is turned off, so that the menubar
        # gets hidden right away next time auto-hide is reenabled.
        if not self.enabled:
            self.sticky = False

    def reconnectToMenus(self):
        connect = self.enabled
        newConnections = []

        menu: QMenu
        for menu in self.menuBar.findChildren(QMenu, options=Qt.FindChildOption.FindDirectChildrenOnly):
            menuRef = weakref.ref(menu)
            wasConnected = menuRef in self.menusConnected

            if connect:
                if not wasConnected:
                    menu.aboutToShow.connect(self.onMenuAboutToShow)
                    menu.aboutToHide.connect(self.onMenuAboutToHide)
                newConnections.append(menuRef)
            elif wasConnected:
                menu.aboutToShow.disconnect(self.onMenuAboutToShow)
                menu.aboutToHide.disconnect(self.onMenuAboutToHide)

        self.menusConnected = newConnections

    @property
    def enabled(self):
        return not settings.prefs.showMenuBar

    @property
    def isHidden(self):
        return self.menuBar.maximumHeight() == 0

    def showMenuBar(self, show):
        if show:
            self.menuBar.setMaximumHeight(self.defaultMaximumHeight)
        else:
            self.menuBar.setMaximumHeight(0)

    def toggle(self):
        assert self.enabled, "did you forget to disconnect AutoHideMenuBar signals?"
        self.sticky = True
        self.showMenuBar(self.isHidden)

    def onMenuAboutToShow(self):
        assert self.enabled, "did you forget to disconnect AutoHideMenuBar signals?"
        self.hideScheduler.stop()
        if self.isHidden:
            self.showMenuBar(True)
            self.sticky = False

    def onMenuAboutToHide(self):
        assert self.enabled, "did you forget to disconnect AutoHideMenuBar signals?"
        self.hideScheduler.start()

    def doScheduledHide(self):
        assert self.enabled, "did you forget to disconnect AutoHideMenuBar signals?"
        if not self.isHidden and not self.sticky:
            self.showMenuBar(False)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations

import logging
from collections.abc import Iterable
from itertools import zip_longest
from collections.abc import Sequence

from gitfourchette.graph.graph import (
    ArcJunction,
    BATCHROW_UNDEF,
    BatchRow,
    Frame,
    Graph,
    KF_INTERVAL,
    Oid,
)
from gitfourchette.graph.graphweaver import GraphWeaver
from gitfourchette.toolbox import Benchmark

logger = logging.getLogger(__name__)


class GraphSplicer:
    def __init__(self, oldGraph: Graph, oldHeads: Iterable[Oid], newHeads: Iterable[Oid]):
        self.done = False
        self.foundEquilibrium = False
        self.equilibriumNewRow = -1
        self.equilibriumOldRow = -1
        self.oldGraphRowOffset = 0

        self.newGraph, self.weaver = GraphWeaver.newGraph()

        self.oldGraph = oldGraph
        self.oldPlayer = oldGraph.startPlayback()

        # Commits that we must see before finding the equilibrium.
        newHeads = set(newHeads)
        oldHeads = set(oldHeads)
        self.newHeads = newHeads
        self.requiredNewCommits = (newHeads - oldHeads)  # heads that appeared
        self.requiredOldCommits = (oldHeads - newHeads)  # heads that disappeared

    def spliceNewCommit(
            self,
            newCommit: Oid,
            parentsOfNewCommit: Sequence[Oid],
            keyframeInterval=KF_INTERVAL
    ):
        assert not self.done

        # Generate arcs for new frame.
        self.weaver.newCommit(newCommit, parentsOfNewCommit)

        # Save keyframe in new context every now and then.
        if int(self.weaver.row) % keyframeInterval == 0:
            self.newGraph.saveKeyframe(self.weaver)

        # Register this commit in the new graph's row sequence.
        self.newGraph.commitRows[newCommit] = self.weaver.row

        # Is it one of the commits that we must see before we can stop consuming new commits?
        if newCommit in self.requiredNewCommits:
            self.requiredNewCommits.remove(newCommit)

        # If the commit wasn't known in the old graph, don't advance the old graph.
        newCommitWasKnown = newCommit in self.oldGraph.commitRows
        if not newCommitWasKnown:
            return

        # The old graph's playback may be positioned past this commit already,
        # e.g. if branches were reordered. In that case, don't advance the old graph.
        if newCommit in self.oldPlayer.seenCommits:
            return

        # We know the new commit is ahead in the old graph.
        # Advance playback of the old graph to the new commit.
        try:
            while self.oldPlayer.commit != newCommit:
                self.oldPlayer.advanceToNextRow()  # May raise StopIteration

                oldCommit = self.oldPlayer.commit
                self.requiredOldCommits.discard(oldCommit)

                # Topological sort may reorder branches. If we've skipped a
                # head that still exists, we've got to see it in the new graph.
                if oldCommit in self.newHeads and oldCommit not in self.newGraph.commitRows:
                    self.requiredNewCommits.add(oldCommit)
                # Otherwise, the old commit is now unreachable. We'll purge it afterward.

            self.requiredOldCommits.discard(newCommit)

        except StopIteration:
            # Old graph depleted.
            self.onGraphDepleted()
            return

        # See if we're done: no more commits we want to see,
        # and the graph frames start being "equal" in both graphs.
        if (not self.requiredNewCommits and
                not self.requiredOldCommits and
                self.isEquilibriumReached(self.weaver, self.oldPlayer)):
            self.onEquilibriumFound()
            return

    def onEquilibriumFound(self):
        """Completion with equilibrium"""

        self.done = True
        self.foundEquilibrium = True

        # We'll basically concatenate newContext[eqNewRow:] and oldContext[:eqOldRow].
        equilibriumNewRow = int(self.weaver.row)
        equilibriumOldRow = int(self.oldPlayer.row)
        rowShiftInOldGraph = equilibriumNewRow - equilibriumOldRow

        # Save rows for use by external code
        self.equilibriumNewRow = equilibriumNewRow
        self.equilibriumOldRow = equilibriumOldRow
        self.oldGraphRowOffset = rowShiftInOldGraph

        logger.debug(f"Equilibrium: commit={self.oldPlayer.commit!s:.7} new={equilibriumNewRow} old={equilibriumOldRow}")

        # We can bail now if nothing changed.
        if equilibriumOldRow == 0 and equilibriumNewRow == 0:
            return

        # After reaching equilibrium there might still be open arcs that aren't closed yet.
        # Let's find out where they end before we can concatenate the graphs.
        equilibriumNewOpenArcs = list(filter(None, self.weaver.openArcs))
        equilibriumOldOpenArcs = list(filter(None, self.oldPlayer.sealCopy().openArcs))
        assert len(equilibriumOldOpenArcs) == len(equilibriumNewOpenArcs)

        # Fix up dangling open arcs in new graph
        for oldOpenArc, newOpenArc in zip(equilibriumOldOpenArcs, equilibriumNewOpenArcs, strict=True):
            # Find out where the arc is resolved
            assert newOpenArc.openedBy == oldOpenArc.openedBy
            assert newOpenArc.closedBy == oldOpenArc.closedBy
            assert newOpenArc.closedAt == BATCHROW_UNDEF  # new graph's been interrupted before resolving this arc
            newOpenArc.closedAt = oldOpenArc.closedAt

            # Remap chain - the ChainHandle object is shared with all arcs on this chain
            newCH = newOpenArc.chain
            oldCH = oldOpenArc.chain
            assert newCH.topRow.isValid()
            assert oldCH.topRow.isValid()
            newCH.bottomRow = oldCH.bottomRow   # rewire bottom row BEFORE setting alias
            oldCH.setAliasOf(newCH)

            # Splice old junctions into new junctions
            if oldOpenArc.junctions:
                junctions: list[ArcJunction] = []
                junctions.extend(j for j in newOpenArc.junctions if j.joinedAt <= equilibriumNewRow)  # before eq
                junctions.extend(j for j in oldOpenArc.junctions if j.joinedAt > equilibriumOldRow)  # after eq
                assert all(junctions.count(x) == 1 for x in junctions), "duplicate junctions after splicing"
                newOpenArc.junctions = junctions

        # Do the actual splicing.

        # If we're adding a commit at the top of the graph, the closed arcs of the first keyframe will be incorrect,
        # so we must make sure to nuke the keyframe for equilibriumOldRow if it exists.
        with Benchmark("Delete lost keyframes"):
            self.oldGraph.deleteKeyframesDependingOnRowsAbove(equilibriumOldRow + 1)

        with Benchmark("Delete lost arcs"):
            self.oldGraph.deleteArcsDependingOnRowsAbove(equilibriumOldRow)

        with Benchmark("Delete lost rows"):
            for lostCommit in (self.oldPlayer.seenCommits - self.newGraph.commitRows.keys()):
                del self.oldGraph.commitRows[lostCommit]

        with Benchmark(F"Shift {len(self.oldGraph.ownBatches)} old batches by {rowShiftInOldGraph} rows"):
            BatchRow.BatchManager.shiftBatches(rowShiftInOldGraph, self.oldGraph.ownBatches)

        with Benchmark("Insert Front"):
            self.oldGraph.insertFront(self.newGraph, equilibriumNewRow)

        with Benchmark("Update row cache"):
            self.oldGraph.commitRows.update(self.newGraph.commitRows)
            self.oldGraph.ownBatches.extend(self.newGraph.ownBatches)  # Steal newGraph's batches
            self.newGraph.ownBatches = []  # Don't let newGraph nuke the batches in its __del__

        # Invalidate volatile player, which may be referring to dead keyframes
        self.oldGraph.volatilePlayer = None

    def onGraphDepleted(self):
        """Completion without equilibrium: no more commits in oldGraph"""

        self.done = True

        # If we exited the loop without reaching equilibrium, the whole graph has changed.
        # In that case, steal the contents of newGraph, and bail.

        self.equilibriumOldRow = int(self.oldPlayer.row)
        self.equilibriumNewRow = int(self.weaver.row)
        self.oldGraphRowOffset = 0

        self.oldGraph.shallowCopyFrom(self.newGraph)

    @staticmethod
    def isEquilibriumReached(frameA: Frame, frameB: Frame):
        rowA = int(frameA.row)
        rowB = int(frameB.row)

        for arcA, arcB in zip_longest(frameA.openArcs, frameB.openArcs):
            isStaleA = (not arcA) or arcA.isStale(rowA)
            isStaleB = (not arcB) or arcB.isStale(rowB)

            if isStaleA != isStaleB:
                return False

            if isStaleA:
                assert isStaleB
                continue

            assert arcA.lane == arcB.lane

            if not (arcA.openedBy == arcB.openedBy and arcA.closedBy == arcB.closedBy):
                return False

        # Do NOT test solved arcs!
        return True
